from typing import List

from openapi_client.models.response_kernel_resource_spec import (
    ResponseKernelResourceSpec,
)
from vessl import vessl_api
from vessl.organization import _get_organization_name
from vessl.util.exception import InvalidKernelResourceSpecError


def read_kernel_resource_spec(
    cluster_id: int, kernel_resource_spec_name: str, **kwargs
) -> ResponseKernelResourceSpec:
    """Read kernel resource spec in the default organization. If you
    want to override the default organization, then pass `organization_name` as
    `**kwargs`.

    Args:
        cluster_id(int): Cluster ID.
        kernel_resource_spec_name(str): Kernel resource spec name.

    Example:
        ```python
        vessl.read_kernel_resource_spec(
            cluster_id=123,
            kernel_resource_spec_name="v1.cpu-4.mem-13",
        )
        ```
    """
    kernel_resource_specs = list_kernel_resource_specs(cluster_id, **kwargs)
    kernel_resource_specs = {x.name: x for x in kernel_resource_specs}

    if kernel_resource_spec_name not in kernel_resource_specs:
        raise InvalidKernelResourceSpecError(
            f"Kernel resource spec not found: {kernel_resource_spec_name}"
        )
    return kernel_resource_specs[kernel_resource_spec_name]


def list_kernel_resource_specs(
    cluster_id: int,
    **kwargs,
) -> List[ResponseKernelResourceSpec]:
    """List kernel resource specs in the default organization. If you
    want to override the default organization, then pass `organization_name` as
    `**kwargs`.

    Args:
        cluster_id(int): Cluster ID.

    Example:
        ```python
        vessl.list_kernel_resource_specs(
            cluster_id=123,
        )
        ```
    """
    return vessl_api.kernel_resource_spec_list_api(
        cluster_id=cluster_id,
        organization_name=_get_organization_name(**kwargs),
    ).results


def _configure_custom_kernel_resource_spec(
    processor_type: str = None,
    cpu_limit: float = None,
    memory_limit: str = None,
    gpu_type: str = None,
    gpu_limit: int = None,
) -> ResponseKernelResourceSpec:
    return ResponseKernelResourceSpec(
        processor_type=processor_type,
        cpu_type="Any",
        cpu_limit=cpu_limit,
        memory_limit=memory_limit,
        gpu_type=gpu_type,
        gpu_limit=gpu_limit,
    )



from typing import List

import click

from openapi_client.models import ResponseKernelClusterInfo
from vessl.cli._base import VesslGroup, vessl_argument, vessl_option
from vessl.cli._util import (
    Endpoint,
    format_string,
    generic_prompter,
    print_data,
    print_table,
    prompt_choices,
)
from vessl.cli.organization import organization_name_option
from vessl.kernel_cluster import (
    _check_cluster_install_dependencies,
    _check_existing_vessl_installation,
    _install_k0s_in_docker,
    _load_cluster_kubeconfig,
    create_cluster,
    delete_cluster,
    list_cluster_nodes,
    list_clusters,
    read_cluster,
    rename_cluster,
)
from vessl.organization import _get_organization_name
from vessl.util.constant import CLUSTER_MODE_SINGLE
from vessl.util.exception import VesslRuntimeException


def cluster_name_prompter(ctx: click.Context, param: click.Parameter, value: str) -> str:
    clusters = list_clusters()
    return prompt_choices("Cluster", [x.name for x in clusters])


def custom_cluster_name_prompter(ctx: click.Context, param: click.Parameter, value: str) -> str:
    clusters = list_clusters()
    clusters = [x for x in clusters if not x.is_savvihub_managed]
    return prompt_choices("Cluster", [x.name for x in clusters])


@click.command(name="cluster", cls=VesslGroup)
def cli():
    pass


@cli.vessl_command()
@vessl_option(
    "--name",
    type=click.STRING,
    required=True,
    help="Name of the cluster. (e.g. 'seoul-cluster')",
)
@vessl_option(
    "--mode",
    type=click.STRING,
    default="single",
    help="Cluster installation mode; single|multinode. If mode=single, VESSL will be installed in a Docker container as a single-node mode.",
)
@vessl_option(
    "--kubernetes-namespace",
    type=click.STRING,
    default="vessl",
    help="Kubernetes namespace to install VESSL agent (default: vessl)",
)
@vessl_option(
    "--provider",
    type=click.STRING,
    default="on-premise",
    help="Cluster provider; aws|gcp|on-premise (default: on-premise)",
)
@vessl_option(
    "--extra-helm-values",
    type=click.STRING,
    multiple=True,
    default=[],
    help="Extra Helm values to pass when installing VESSL agent.",
)
@organization_name_option
def create(
    name: str, mode: str, kubernetes_namespace: str, provider: str, extra_helm_values: List[str]
):
    _check_cluster_install_dependencies(mode)
    loaded, (config_path, context) = _load_cluster_kubeconfig(mode)
    if not loaded:
        if mode == CLUSTER_MODE_SINGLE:
            click.confirm(
                f"Kubernetes cluster not found; install kubernetes-in-docker on the machine?",
                abort=True,
            )
            config_path, context = _install_k0s_in_docker()
        else:
            raise VesslRuntimeException(
                "Kubernetes cluster not found; Please configure valid kubeconfig file location by KUBECONFIG envvar or install Kubernetes on the machine."
            )

    vessl_installation_info = _check_existing_vessl_installation()
    if vessl_installation_info is not None:
        print_cluster_data(vessl_installation_info)
        return

    if not provider:
        provider = "on-premise"

    input_data = {
        "VESSL cluster name": name,
        "Provider": provider,
        "Kubeconfig path": config_path,
        "Current kubernetes context": context,
        "Kubernetes namespace": kubernetes_namespace,
    }
    if extra_helm_values and len(extra_helm_values) > 0:
        input_data["Extra Helm values"] = ", ".join(extra_helm_values)

    print_data(input_data)
    click.confirm(
        f"Are you sure you want to install VESSL agent on current Kubernetes cluster?",
        abort=True,
    )
    vessl_installation_info = create_cluster(
        cluster_name=name,
        kubernetes_namespace=kubernetes_namespace,
        kubeconfig_path=config_path,
        provider=provider,
        extra_helm_values=extra_helm_values,
    )
    print(f"cluster {name} created!")
    print_cluster_data(vessl_installation_info)


@cli.vessl_command()
@vessl_argument("name", type=click.STRING, required=True, prompter=cluster_name_prompter)
@organization_name_option
def read(name: str):
    cluster = read_cluster(cluster_name=name)
    print_cluster_data(cluster)


@cli.vessl_command()
@organization_name_option
def list():
    clusters = list_clusters()
    print_table(
        clusters,
        ["ID", "Name", "Type", "Status", "K8s Master Endpoint", "K8s Namespace"],
        lambda x: [
            x.id,
            x.name,
            "Managed" if x.is_savvihub_managed else "Custom",
            x.status.replace("-", " "),
            format_string(x.name),
            format_string(x.kubernetes_namespace),
        ],
    )


@cli.vessl_command()
@vessl_argument("name", type=click.STRING, required=True, prompter=custom_cluster_name_prompter)
@organization_name_option
def delete(name: str):
    click.confirm(f"Are you sure you want to delete '{name}'?", abort=True)
    cluster = read_cluster(cluster_name=name)
    data = delete_cluster(cluster_id=cluster.id)
    print(f"Deleted '{name}'.")
    print(f"Note that this command does NOT destroy your kubernetes cluster.")
    print(f"If you want to destroy the cluster installed with VESSL CLI,")
    print(f"try execute the following in the shell (note that this is NOT RECOVERABLE).")
    print(f"$ docker stop vessl-k0s && docker rm vessl-k0s")
    print(
        f'$ docker run -it --rm --privileged --pid=host alpine nsenter -t 1 -m -- sh -c "rm -rfv /var/lib/k0s"'
    )


@cli.vessl_command()
@vessl_argument("name", type=click.STRING, required=True, prompter=custom_cluster_name_prompter)
@vessl_argument("new_name", type=click.STRING, required=True, prompter=generic_prompter("New name"))
@organization_name_option
def rename(name: str, new_name: str):
    cluster = read_cluster(cluster_name=name)
    cluster = rename_cluster(cluster_id=cluster.id, new_cluster_name=new_name)
    print(f"Renamed '{name}' to '{cluster.name}'.")


@cli.vessl_command()
@vessl_argument("name", type=click.STRING, required=True, prompter=custom_cluster_name_prompter)
@organization_name_option
def list_nodes(name: str):
    cluster = read_cluster(cluster_name=name)
    cluster_nodes = list_cluster_nodes(cluster_id=cluster.id)
    print_table(
        cluster_nodes,
        ["ID", "Name", "CPU Limits", "GPU Limits", "Memory Limits"],
        lambda x: [x.id, x.name, x.cpu_limits, x.gpu_limits, x.memory_limits],
    )


def cluster_id_prompter(ctx: click.Context, param: click.Parameter, value: str) -> int:
    clusters = list_clusters()
    cluster = prompt_choices(
        "Cluster",
        [(f"{x.name}", x) for x in clusters],
    )
    ctx.obj["cluster"] = cluster
    return cluster.id


def cluster_name_prompter(ctx: click.Context, param: click.Parameter, value: str) -> str:
    clusters = list_clusters()
    cluster = prompt_choices(
        "Cluster",
        [(f"{x.name}", x) for x in clusters],
    )
    ctx.obj["cluster"] = cluster
    return cluster.name


def cluster_name_callback(ctx: click.Context, param: click.Parameter, value: str) -> str:
    if "cluster" not in ctx.obj:
        ctx.obj["cluster"] = read_cluster(value)
    return value


def print_cluster_data(cluster: ResponseKernelClusterInfo, **kwargs):
    print_data(
        {
            "Name": cluster.name,
            "Dashboard URL": Endpoint.cluster.format(_get_organization_name(**kwargs), cluster.id),
            "Type": "VESSL-managed" if cluster.is_savvihub_managed else "Custom",
            "Region": format_string(cluster.region),
            "Status": cluster.status.replace("-", " "),
            "K8s Master Endpoint": format_string(cluster.name),
            "K8s Namespace": format_string(cluster.kubernetes_namespace),
            "K8s Service Type": cluster.kubernetes_service_type,
        }
    )


cluster_option = vessl_option(
    "-c",
    "--cluster",
    type=click.STRING,
    required=True,
    prompter=cluster_name_prompter,
    callback=cluster_name_callback,
    help="Must be specified before resource-related options (`--resource`, `--processor`, ...).",
)



import os
from typing import Optional, Union

import six

from vessl.util import logger
from vessl.util.common import generate_uuid, get_module
from vessl.util.constant import VESSL_MEDIA_PATH


class Image:
    def __init__(
        self,
        data: Union[str, "Image", "PIL.Image", "numpy.ndarray"],
        caption: Optional[str] = "",
        mode: Optional[str] = None,
    ):
        self._image = None
        self._caption = None
        self._mode = None
        self._path = None

        self._init_meta(caption, mode)
        self._init_image(data)
        self._save_image()

    def _init_meta(self, caption, mode):
        self._caption = caption
        self._mode = mode

        os.makedirs(VESSL_MEDIA_PATH, exist_ok=True)
        self._path = os.path.join(VESSL_MEDIA_PATH, generate_uuid() + ".png")

    def _init_image_from_path(self, data):
        pil_image = get_module(
            "PIL.Image",
            required='Pillow package is required. Run "pip install Pillow".',
        )
        self._image = pil_image.open(data)

    def _init_image_from_data(self, data):
        pil_image = get_module(
            "PIL.Image",
            required='Pillow package is required. Run "pip install Pillow".',
        )
        if Image.get_type_name(data).startswith("torch."):
            self._image = pil_image.fromarray(
                data.mul(255).clamp(0, 255).byte().permute(1, 2, 0).squeeze().cpu().numpy()
            )
        elif isinstance(data, pil_image.Image):
            self._image = data
        else:
            if hasattr(data, "numpy"):
                data = data.numpy()
            if data.ndim > 2:
                data = data.squeeze()
            self._image = pil_image.fromarray(
                Image.to_uint8(data),
                mode=self._mode,
            )

    def _init_image(self, data):
        if isinstance(data, Image):
            self._image = data._image
        elif isinstance(data, six.string_types):
            self._init_image_from_path(data)
        else:
            self._init_image_from_data(data)

    def _save_image(self):
        self._image.save(self._path)

    def flush(self):
        if os.path.isfile(self._path):
            os.remove(self._path)
        else:
            logger.error(f"Error: {self._path} file not found")

    @property
    def path(self):
        return self._path

    @property
    def caption(self):
        return self._caption

    @classmethod
    def get_type_name(cls, obj):
        type_name = obj.__class__.__module__ + "." + obj.__class__.__name__
        if type_name in ["builtins.module", "__builtin__.module"]:
            return obj.__name__
        else:
            return type_name

    @classmethod
    def to_uint8(cls, data):
        np = get_module(
            "numpy",
            required='numpy package is required. Run "pip install numpy"',
        )
        dmin = np.min(data)
        if dmin < 0:
            data = (data - np.min(data)) / np.ptp(data)
        if np.max(data) <= 1.0:
            data = (data * 255).astype(np.int32)

        return data.clip(0, 255).astype(np.uint8)



from typing import List

from openapi_client import ResponseOrganizationCredentialsInfo
from openapi_client.models import (
    OrganizationCreateAPIInput,
    ResponseOrganization,
    ResponseOrganizationInfo,
)
from vessl import vessl_api
from vessl.util.exception import InvalidOrganizationError


def read_organization(organization_name: str) -> ResponseOrganizationInfo:
    """Read organization

    Args:
        organization_name(str): Organization name.

    Example:
        ```python
        vessl.read_organization(
            organization_name="foo"
        )
        ```
    """
    return vessl_api.organization_read_api(organization_name=organization_name)


def list_organizations() -> List[ResponseOrganization]:
    """List organizations

    Example:
        ```python
        vessl.list_organizations()
        ```
    """
    return vessl_api.organization_list_api().organizations


def create_organization(
    organization_name: str,
) -> ResponseOrganizationInfo:
    """Create organization

    Args:
        organization_name(str): Organization name.

    Example:
        ```python
        vessl.create_organization(
            organization_name="foo",
        )
        ```
    """
    return vessl_api.organization_create_api(
        organization_create_api_input=OrganizationCreateAPIInput(
            name=organization_name,
        )
    )


def list_organization_credentials(
    types: [str],
    components: [str],
    **kwargs,
) -> ResponseOrganizationCredentialsInfo:
    """List organization credentials

    Args:
        types([str]): list of credential types.
        components([str]): list of credential components.

    Example:
        ```python
        vessl.list_organization_credentials(
            organization_name="foo",
            types= ["aws-access-key", "docker-credentials"],
            components: ["image"],
        )
        ```
    """
    return vessl_api.organization_credentials_list_api(
        organization_name=_get_organization_name(**kwargs),
        types=types,
        components=components,
    ).results


def _get_organization_name(**kwargs) -> str:
    organization_name = kwargs.get("organization_name")
    if organization_name is not None:
        return organization_name
    if vessl_api.organization is not None:
        return vessl_api.organization.name
    raise InvalidOrganizationError("No organization selected.")



import os
import subprocess
import urllib.parse
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List

import click
import humanfriendly
import inquirer
import paramiko

from openapi_client import CliWorkspaceBackupCreateAPIInput
from openapi_client.models import (
    InfluxdbWorkloadLog,
    OrmVolumeMountRequests,
    OrmWorkspacePort,
    ResponseWorkspaceDetail,
    ResponseWorkspaceList,
    WorkspaceCreateAPIInput,
)
from vessl import vessl_api
from vessl.kernel_cluster import list_cluster_nodes, read_cluster
from vessl.kernel_resource_spec import (
    _configure_custom_kernel_resource_spec,
    read_kernel_resource_spec,
)
from vessl.organization import _get_organization_name
from vessl.util.common import parse_time_to_ago
from vessl.util.config import VesslConfigLoader
from vessl.util.constant import (
    SSH_CONFIG_FORMAT,
    SSH_CONFIG_PATH,
    TEMP_DIR,
    WORKSPACE_BACKUP_MAX_SIZE,
    WORKSPACE_BACKUP_MAX_SIZE_FORMATTED,
)
from vessl.util.exception import InvalidWorkspaceError, VesslException
from vessl.util.random import random_string
from vessl.util.tar import Tar
from vessl.util.zipper import Zipper
from vessl.volume import (
    _configure_volume_mount_request_datasets,
    _configure_volume_mount_request_local_files,
    copy_volume_file,
)


def read_workspace(workspace_id: int, **kwargs) -> ResponseWorkspaceDetail:
    """Read workspace in the default organization. If you want to override the
    default organization, then pass `organization_name` as `**kwargs`.

    Args:
        workspace_id(int): Workspace ID.

    Example:
        ```python
        vessl.read_workspace(
            workspace_id=123456,
        )
        ```
    """
    return vessl_api.workspace_read_api(
        workspace_id=workspace_id, organization_name=_get_organization_name(**kwargs)
    )


def list_workspaces(
    cluster_id: int = None, statuses: List[str] = None, mine: bool = True, **kwargs
) -> List[ResponseWorkspaceList]:
    """List workspaces in the default organization. If you want to override the
    default organization, then pass `organization_name` as `**kwargs`.

    Args:
        cluster_id(int): Defaults to None.
        statuses(List[str]): A list of status filter. Defaults to None.
        mine(bool): True if list only my workspaces, False otherwise. Defaults
            to True.

    Example:
        ```python
        vessl.list_workspaces(
            cluster_id=123456,
            statuses=["running"],
        )
        ```
    """
    statuses = (
        [",".join(statuses)] if statuses else None
    )  # since openapi-generator uses repeating params instead of commas
    return vessl_api.workspace_list_api(
        organization_name=_get_organization_name(**kwargs),
        cluster=cluster_id,
        mine=mine,
        statuses=statuses,
    ).results


def create_workspace(
    name: str,
    cluster_name: str,
    cluster_node_names: List[str] = None,
    kernel_resource_spec_name: str = None,
    processor_type: str = None,
    cpu_limit: float = None,
    memory_limit: str = None,
    gpu_type: str = None,
    gpu_limit: int = None,
    kernel_image_url: str = None,
    max_hours: int = 24,
    dataset_mounts: List[str] = None,
    local_files: List[str] = None,
    root_volume_size: str = "100Gi",
    ports: List[Dict[str, Any]] = None,
    init_script: str = None,
    **kwargs,
) -> ResponseWorkspaceDetail:
    """Create workspace in the default organization. If you want to override the
    default organization, then pass `organization_name` as `**kwargs`.

    Args:
        name(str): Workspace name.
        cluster_name(str): Cluster name(must be specified before other options).
        cluster_node_names(List[str]): A list of candidate cluster node names.
            Defaults to None.
        kernel_resource_spec_name(str):  Resource type to run an experiment (for
            managed cluster only). Defaults to None.
        processor_type(str) cpu or gpu (for custom cluster only). Defaults to
            None.
        cpu_limit(float): Number of vCPUs (for custom cluster only). Defaults to
            None.
        memory_limit(str): Memory limit in GiB (for custom cluster only).
            Defaults to None.
        gpu_type(str): GPU type (for custom cluster only). Defaults to None.
        gpu_limit(int): Number of GPU cores (for custom cluster only). Defaults
            to None.
        kernel_image_url(str): Kernel docker image URL. Defaults to None.
        max_hours(int): Max hours limit to run. Defaults to 24.
        dataset_mounts(List[str]): A list of dataset mounts. Defaults to None.
        local_files(List[str]): A list of local file mounts. Defaults to None.
        root_volume_size(str): Root volume size. Defaults to "100Gi".
        ports(List[Dict[str, Any]]): Port numbers to expose. Defaults to None.
        init_script(str) Custom init script. Defaults to None.

    Example:
        ```python
        vessl.create_workspace(
            name="modern-kick",
            cluster_name="aws-apne2-prod1",
            kernel_resource_spec_name="v1.cpu-0.mem-1",
            kernel_image_url="public.ecr.aws/vessl/kernels:py36.full-cpu.jupyter",
        )
        ```
    """
    cluster = read_cluster(cluster_name)
    cluster_node_ids = (
        [n.id for n in list_cluster_nodes(cluster.id, **kwargs) if n.name in cluster_node_names]
        if cluster_node_names is not None
        else None
    )

    kernel_resource_spec = kernel_resource_spec_id = None
    if kernel_resource_spec_name:
        kernel_resource_spec_id = read_kernel_resource_spec(
            cluster.id,
            kernel_resource_spec_name,
        ).id
    else:
        kernel_resource_spec = _configure_custom_kernel_resource_spec(
            processor_type,
            cpu_limit,
            memory_limit,
            gpu_type,
            gpu_limit,
        )

    all_ports = [
        OrmWorkspacePort(expose_type="http", port=8888, name="jupyter"),
        OrmWorkspacePort(expose_type="tcp", port=22, name="ssh"),
    ]
    if ports is not None:
        for p in ports:
            all_ports.append(
                OrmWorkspacePort(
                    expose_type=p.get("expose_type"),
                    port=p.get("port"),
                    name=p.get("name"),
                )
            )

    requests = []
    if dataset_mounts is not None:
        requests.extend(
            _configure_volume_mount_request_datasets(
                dataset_mounts, should_add_project_datasets=False, **kwargs
            )
        )
    if local_files:
        requests.extend(_configure_volume_mount_request_local_files(local_files, **kwargs))

    volume_mount_requests = OrmVolumeMountRequests(
        root_volume_size=root_volume_size,
        requests=requests,
    )

    return vessl_api.workspace_create_api(
        organization_name=_get_organization_name(**kwargs),
        workspace_create_api_input=WorkspaceCreateAPIInput(
            cluster_id=cluster.id,
            cluster_node_ids=cluster_node_ids,
            resource_spec=kernel_resource_spec,
            resource_spec_id=kernel_resource_spec_id,
            image_url=kernel_image_url,
            name=name,
            max_hours=max_hours,
            ports=all_ports,
            volumes=volume_mount_requests,
            init_script=init_script,
        ),
    )


def list_workspace_logs(
    workspace_id: int,
    tail: int = 200,
    **kwargs,
) -> List[InfluxdbWorkloadLog]:
    """List experiment logs in the default organization. If you want to override
    the default organization, then pass `organization_name` as `**kwargs`.

    Args:
        workspace_id(int): Workspace ID.
        tail (int): The number of lines to display from the end. Display all if
            -1. Defaults to 200.

    Example:
        ```python
         vessl.list_workspace_logs(
            workspace_id=123456,
         )
        ```
    """
    if tail == -1:
        tail = None

    return vessl_api.workspace_logs_api(
        workspace_id=workspace_id,
        organization_name=_get_organization_name(**kwargs),
        limit=tail,
    ).logs


def start_workspace(workspace_id: int, **kwargs) -> ResponseWorkspaceDetail:
    """Start the workspace container in the default organization. If you want to
    override the default organization, then pass `organization_name` as
    `**kwargs`.

    Args:
        workspace_id(int): Workspace ID.

    Example:
        ```python
        vessl.start_workspace(
            workspace_id=123456,
        )
        ```
    """
    return vessl_api.workspace_start_api(
        organization_name=_get_organization_name(**kwargs), workspace_id=workspace_id
    )


def stop_workspace(workspace_id: int, **kwargs) -> ResponseWorkspaceDetail:
    """Stop the workspace container in the default organization. If you want to
    override the default organization, then pass `organization_name` as
    `**kwargs`.

    Args:
        workspace_id(int): Workspace ID.

    Example:
        ```python
        vessl.stop_workspace(
            workspace_id=123456,
        )
        ```
    """
    return vessl_api.workspace_stop_api(
        organization_name=_get_organization_name(**kwargs), workspace_id=workspace_id
    )


def terminate_workspace(workspace_id: int, **kwargs) -> ResponseWorkspaceDetail:
    """Terminate the workspace container in the default organization. If you
    want to override the default organization, then pass `organization_name` as
    `**kwargs`.

    Args:
        workspace_id(int): Workspace ID.

    Example:
        ```python
        vessl.terminate_workspace(
            workspace_id=123456,
        )
        ```
    """
    return vessl_api.workspace_terminate_api(
        organization_name=_get_organization_name(**kwargs), workspace_id=workspace_id
    )


def backup_workspace() -> None:
    """Backup the home directory of the workspace. This command should be called
    inside a workspace.

    Example:
        ```python
        vessl.backup_workspace()
        ```
    """
    workspace_id = VesslConfigLoader().workspace
    if workspace_id is None:
        raise InvalidWorkspaceError("Can only be called within a workspace.")

    workspace = read_workspace(workspace_id)

    filename = (
        f'workspace-backup-{datetime.utcnow().strftime("%Y%m%d%H%M%S")}-{random_string()}.tar.gz'
    )
    gzip_file_full_path = os.path.join(TEMP_DIR, filename)
    home_dir = str(Path.home())

    home_dir_size = -1
    try:
        # use -k option because there is no -b option on UNIX du
        home_dir_size_in_kilobytes = int(
            subprocess.check_output(["du", "-sk", home_dir]).decode().split("\t")[0]
        )
        home_dir_size = home_dir_size_in_kilobytes * 1024
    except Exception:
        pass

    if home_dir_size > WORKSPACE_BACKUP_MAX_SIZE:
        formatted_size = humanfriendly.format_size(home_dir_size, binary=True)
        raise VesslException(
            f"Failed: The size of {home_dir} should be less than {WORKSPACE_BACKUP_MAX_SIZE_FORMATTED} to backup. "
            f"Current size is {formatted_size}.",
        )

    print("Creating a backup file...")
    size = Tar.gzip(gzip_file_full_path, home_dir, exclude_paths=[".vessl"])
    if size > WORKSPACE_BACKUP_MAX_SIZE:
        formatted_size = humanfriendly.format_size(size, binary=True)
        raise VesslException(
            f"Failed: The size of {home_dir} should be less than {WORKSPACE_BACKUP_MAX_SIZE_FORMATTED} to backup. "
            f"Current size is {formatted_size}.",
        )

    print(f"Backup file size: {humanfriendly.format_size(size, binary=True)}")
    print("Uploading the backup file...")
    copy_volume_file(
        source_volume_id=None,
        source_path=gzip_file_full_path,
        dest_volume_id=workspace.backup_volume_id,
        dest_path=filename,
    )
    os.remove(gzip_file_full_path)

    vessl_api.cli_workspace_backup_create_api(
        workspace_id=workspace_id,
        cli_workspace_backup_create_api_input=CliWorkspaceBackupCreateAPIInput(
            filename=filename,
        ),
    )


def restore_workspace() -> None:
    """Restore the home directory from the previous backup. This command should
    be called inside a workspace.

    Example:
        ```python
        vessl.restore_workspace()
        ```
    """
    workspace_id = VesslConfigLoader().workspace
    if workspace_id is None:
        raise InvalidWorkspaceError("Can only be called within a workspace.")

    workspace = read_workspace(workspace_id=workspace_id)
    if workspace.last_backup is None:
        raise click.ClickException("This workspace does not have any backup.")

    if workspace.last_backup.filename.endswith(".zip"):
        dest_path = os.path.join(TEMP_DIR, "workspace-backup.zip")
        print("Downloading the backup file...")
        copy_volume_file(
            source_volume_id=workspace.backup_volume_id,
            source_path=workspace.last_backup.filename,
            dest_volume_id=None,
            dest_path=dest_path,
        )

        zipper = Zipper(dest_path, "r")
        size = zipper.size()
        print(f"Backup file size: {size}")
        extract_path = str(Path.home())
        if size > WORKSPACE_BACKUP_MAX_SIZE:
            extract_path = os.path.join(TEMP_DIR, "workspace-backup")
        print("Extracting...")
        zipper.extractall(extract_path)
        if size > WORKSPACE_BACKUP_MAX_SIZE:
            print(
                f"Restored to {extract_path} "
                f"since the size of the backup file is larger than {WORKSPACE_BACKUP_MAX_SIZE_FORMATTED}. "
            )
        zipper.close()
        zipper.remove()
    else:
        gzip_file_full_path = os.path.join(TEMP_DIR, "workspace-backup.tar.gz")
        print("Downloading the backup file...")
        copy_volume_file(
            source_volume_id=workspace.backup_volume_id,
            source_path=workspace.last_backup.filename,
            dest_volume_id=None,
            dest_path=gzip_file_full_path,
        )

        print("Extracting...")
        Tar.extract(gzip_file_full_path, str(Path.home()))
        os.remove(gzip_file_full_path)


def connect_workspace_ssh(private_key_path: str) -> None:
    """Connect to a running workspace via SSH.

    Args:
        private_key_path(str): SSH private key path

    Example:
        ```python
        vessl.connect_workspace_ssh(
            private_key_path="~/.ssh/key_path",
        )
        ```
    """
    running_workspaces = list_workspaces(statuses=["running"], mine=True)
    if len(running_workspaces) == 0:
        raise click.ClickException("There is no running workspace.")

    if len(running_workspaces) == 1:
        workspace = running_workspaces[0]
    else:
        workspace = inquirer.prompt(
            [
                inquirer.List(
                    "question",
                    message="Select workspace",
                    choices=[
                        (f"{w.name} (created {parse_time_to_ago(w.created_dt)})", w)
                        for w in running_workspaces
                    ],
                )
            ],
            raise_keyboard_interrupt=True,
        ).get("question")

    ssh_endpoint = urllib.parse.urlparse(workspace.endpoints.ssh.endpoint)
    ssh_private_key_option = f" -i {private_key_path}" if private_key_path else ""
    os.system(
        f"ssh -p {ssh_endpoint.port}{ssh_private_key_option} " f"root@{ssh_endpoint.hostname}"
    )


def update_vscode_remote_ssh(private_key_path: str) -> None:
    """Update .ssh/config file for VSCode Remote-SSH plugin.

    Args:
        private_key_path(str): SSH private key path

    Example:
        ```python
        vessl.update_vscode_remote_ssh(
            private_key_path="~/.ssh/key_path",
        )
        ```
    """
    running_workspaces = list_workspaces(statuses=["running"], mine=True)
    if len(running_workspaces) == 0:
        raise click.ClickException("There is no running workspace.")

    ssh_config = paramiko.SSHConfig()
    try:
        with open(SSH_CONFIG_PATH, "r") as fr:
            ssh_config.parse(fr)
    except FileNotFoundError:
        pass

    host_set = ssh_config.get_hostnames()
    for workspace in running_workspaces:
        host = f"{workspace.name}-{int(workspace.created_dt.timestamp())}"
        if host in host_set:
            continue

        ssh_endpoint = urllib.parse.urlparse(workspace.endpoints.ssh.endpoint)

        config_value = SSH_CONFIG_FORMAT.format(
            host=host,
            hostname=ssh_endpoint.hostname,
            port=ssh_endpoint.port,
        )
        if private_key_path:
            config_value += f"    IdentityFile {private_key_path}\n"

        with open(SSH_CONFIG_PATH, "a") as f:
            f.write(config_value)



import os
from test.conftest import USE_MOCK

import pytest
from mock import Mock, patch

import vessl
from vessl.util.config import VesslConfigLoader
from vessl.util.constant import (
    ACCESS_TOKEN_ENV_VAR,
    CREDENTIALS_FILE_ENV_VAR,
    DEFAULT_ORGANIZATION_ENV_VAR,
)
from vessl.util.exception import (
    InvalidOrganizationError,
    InvalidProjectError,
    InvalidTokenError,
    VesslApiException,
)

config_path_api = "test/config_api"


def teardown_module(module):
    try:
        os.remove(config_path_api)
    except OSError:
        pass

    # Reset access token and organization because tests override it
    if not USE_MOCK:
        vessl.init(is_test=True)


def test_configure_access_token():
    config_loader = VesslConfigLoader(config_path_api)
    config_loader_ = patch("vessl.vessl_api.config_loader", config_loader)
    get_my_user_info_api_ = patch("vessl.vessl_api.get_my_user_info_api", return_value=Mock())

    # Test `access_token`
    access_token = "abcde"
    with config_loader_, get_my_user_info_api_:
        vessl.configure_access_token(access_token=access_token)

    assert vessl.vessl_api.api_client.default_headers["Authorization"] == f"Token {access_token}"
    assert config_loader.access_token == access_token

    # Test `credentials_file`
    access_token = "bcdef"
    config_loader.access_token = access_token
    with config_loader_, get_my_user_info_api_:
        vessl.configure_access_token(credentials_file=config_path_api)

    assert vessl.vessl_api.api_client.default_headers["Authorization"] == f"Token {access_token}"
    assert VesslConfigLoader(config_path_api).access_token == access_token

    # Test `AUTH_TOKEN_ENV_VAR`
    access_token = "cdefg"

    access_token_env_var_ = patch.dict(os.environ, {ACCESS_TOKEN_ENV_VAR: access_token})
    with config_loader_, get_my_user_info_api_, access_token_env_var_:
        vessl.configure_access_token()

    assert vessl.vessl_api.api_client.default_headers["Authorization"] == f"Token {access_token}"
    assert VesslConfigLoader(config_path_api).access_token == access_token

    # Test `CREDENTIALS_FILE_ENV_VAR`
    access_token = "defgh"
    config_loader.access_token = access_token

    credentials_file_env_var_ = patch.dict(os.environ, {CREDENTIALS_FILE_ENV_VAR: config_path_api})
    with config_loader_, get_my_user_info_api_, credentials_file_env_var_:
        vessl.configure_access_token()

    assert vessl.vessl_api.api_client.default_headers["Authorization"] == f"Token {access_token}"
    assert VesslConfigLoader(config_path_api).access_token == access_token

    # Test `force_update`
    access_token = "efghi"
    get_new_access_token_ = patch(
        "vessl.util.api.VesslApi._get_new_access_token", return_value=access_token
    )
    with config_loader_, get_my_user_info_api_, get_new_access_token_:
        vessl.configure_access_token(force_update=True)

    assert vessl.vessl_api.api_client.default_headers["Authorization"] == f"Token {access_token}"
    assert VesslConfigLoader(config_path_api).access_token == access_token

    # Test fail: expired
    get_my_user_info_api_ = patch(
        "vessl.vessl_api.get_my_user_info_api", side_effect=VesslApiException()
    )
    with config_loader_, get_my_user_info_api_:
        with pytest.raises(InvalidTokenError):
            vessl.configure_access_token(access_token=access_token)


def test_configure_organization():
    config_loader = VesslConfigLoader(config_path_api)
    organization = Mock()
    response = Mock()
    response.organizations = [organization]

    config_loader_ = patch("vessl.vessl_api.config_loader", config_loader)
    organization_read_api_ = patch("vessl.vessl_api.organization_read_api", return_value=organization)
    organization_list_api_ = patch("vessl.vessl_api.organization_list_api", return_value=response)

    # Test `organization_name`
    organization_name = "org1"
    organization.name = organization_name

    with config_loader_, organization_read_api_, organization_list_api_:
        vessl.configure_organization(organization_name=organization_name)

    assert vessl.vessl_api.organization.name == organization_name
    assert VesslConfigLoader(config_path_api).default_organization == organization_name

    # Test `credentials_file`
    organization_name = "org2"
    organization.name = organization_name
    config_loader.default_organization = organization_name

    with config_loader_, organization_read_api_, organization_list_api_:
        vessl.configure_organization(credentials_file=config_path_api)

    assert vessl.vessl_api.organization.name == organization_name
    assert VesslConfigLoader(config_path_api).default_organization == organization_name

    # Test `DEFAULT_ORGANIZATION_ENV_VAR`
    organization_name = "org3"
    organization.name = organization_name

    default_organization_env_var_ = patch.dict(
        os.environ, {DEFAULT_ORGANIZATION_ENV_VAR: organization_name}
    )

    with config_loader_, organization_read_api_, organization_list_api_, default_organization_env_var_:
        vessl.configure_organization()

    assert vessl.vessl_api.organization.name == organization_name
    assert VesslConfigLoader(config_path_api).default_organization == organization_name

    # Test `CREDENTIALS_FILE_ENV_VAR`
    organization_name = "org4"
    organization.name = organization_name
    config_loader.default_organization = organization_name

    credentials_file_env_var_ = patch.dict(os.environ, {CREDENTIALS_FILE_ENV_VAR: config_path_api})

    with config_loader_, organization_read_api_, organization_list_api_, credentials_file_env_var_:
        vessl.configure_organization()

    assert vessl.vessl_api.organization.name == organization_name
    assert VesslConfigLoader(config_path_api).default_organization == organization_name

    organization_read_api_ = patch("vessl.vessl_api.organization_read_api", side_effect=VesslApiException)
    # Test fail: no organizations
    with organization_read_api_, organization_list_api_:
        with pytest.raises(InvalidOrganizationError):
            vessl.configure_organization(organization_name=organization_name)

    # Test fail: no such organization
    with organization_read_api_, organization_list_api_:
        with pytest.raises(InvalidOrganizationError):
            vessl.configure_organization(organization_name="nonexistent_organization")


def test_configure_project():
    config_loader = VesslConfigLoader(config_path_api)
    project = Mock()
    response = Mock()
    response.results = [project]

    config_loader_ = patch("vessl.vessl_api.config_loader", config_loader)
    project_list_api_ = patch("vessl.vessl_api.project_list_api", return_value=response)

    # Test `project_name`
    project_name = "proj1"
    project.name = project_name

    with config_loader_, project_list_api_:
        vessl.configure_project(project_name=project_name)

    assert vessl.vessl_api.project.name == project_name
    assert VesslConfigLoader(config_path_api).default_project == project_name

    # Test `credentials_file`
    project_name = "proj2"
    project.name = project_name
    config_loader.default_project = project_name

    with config_loader_, project_list_api_:
        vessl.configure_project(credentials_file=config_path_api)

    assert vessl.vessl_api.project.name == project_name
    assert VesslConfigLoader(config_path_api).default_project == project_name

    # Test `CREDENTIALS_FILE_ENV_VAR`
    project_name = "proj3"
    project.name = project_name
    config_loader.default_project = project_name

    credentials_file_env_var_ = patch.dict(os.environ, {CREDENTIALS_FILE_ENV_VAR: config_path_api})

    with config_loader_, project_list_api_, credentials_file_env_var_:
        vessl.configure_project()

    assert vessl.vessl_api.project.name == project_name
    assert VesslConfigLoader(config_path_api).default_project == project_name

    # Test fail: no projects
    response.results = []
    with project_list_api_:
        with pytest.raises(InvalidProjectError):
            vessl.configure_project(project_name=project_name)

    # Test fail: no such project
    response.results = [Mock()]
    with project_list_api_:
        with pytest.raises(InvalidProjectError):
            vessl.configure_project(project_name="nonexistent_project")



# coding: utf-8

# flake8: noqa
"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

# import models into model package
from openapi_client.models.account_get_my_workload_list_response import AccountGetMyWorkloadListResponse
from openapi_client.models.account_invitation_token_validate_response import AccountInvitationTokenValidateResponse
from openapi_client.models.account_ssh_key_list_response import AccountSSHKeyListResponse
from openapi_client.models.account_sign_in_cli_check_response import AccountSignInCliCheckResponse
from openapi_client.models.account_sign_in_cli_token_response import AccountSignInCliTokenResponse
from openapi_client.models.add_git_ssh_key_api_input import AddGitSSHKeyAPIInput
from openapi_client.models.auto_top_up_config_update_api_input import AutoTopUpConfigUpdateAPIInput
from openapi_client.models.billing_customer_portal_response import BillingCustomerPortalResponse
from openapi_client.models.billing_manual_top_up_response import BillingManualTopUpResponse
from openapi_client.models.bit_bucket_authorize_api_input import BitBucketAuthorizeAPIInput
from openapi_client.models.change_password_api_input import ChangePasswordAPIInput
from openapi_client.models.cli_experiment_freeze_dataset_version_api_input import CliExperimentFreezeDatasetVersionAPIInput
from openapi_client.models.cli_workspace_backup_create_api_input import CliWorkspaceBackupCreateAPIInput
from openapi_client.models.cluster_access_delete_api_input import ClusterAccessDeleteAPIInput
from openapi_client.models.cluster_access_upsert_api_input import ClusterAccessUpsertAPIInput
from openapi_client.models.cluster_cluster_access_list_response import ClusterClusterAccessListResponse
from openapi_client.models.cluster_cluster_list_response import ClusterClusterListResponse
from openapi_client.models.cluster_cluster_node_list_response import ClusterClusterNodeListResponse
from openapi_client.models.cluster_cluster_quota_list_response import ClusterClusterQuotaListResponse
from openapi_client.models.cluster_cluster_quota_upsert_detail import ClusterClusterQuotaUpsertDetail
from openapi_client.models.cluster_cluster_quota_usage_response import ClusterClusterQuotaUsageResponse
from openapi_client.models.cluster_cluster_usage_report_response import ClusterClusterUsageReportResponse
from openapi_client.models.cluster_cluster_workload import ClusterClusterWorkload
from openapi_client.models.cluster_cluster_workload_list_response import ClusterClusterWorkloadListResponse
from openapi_client.models.cluster_cluster_workload_list_with_prometheus_metric_response import ClusterClusterWorkloadListWithPrometheusMetricResponse
from openapi_client.models.cluster_cluster_workload_with_prometheus_metric import ClusterClusterWorkloadWithPrometheusMetric
from openapi_client.models.cluster_custom_cluster_key_response import ClusterCustomClusterKeyResponse
from openapi_client.models.cluster_custom_cluster_node_list_response import ClusterCustomClusterNodeListResponse
from openapi_client.models.cluster_managed_cluster import ClusterManagedCluster
from openapi_client.models.cluster_managed_cluster_list_response import ClusterManagedClusterListResponse
from openapi_client.models.cluster_quota_delete_api_input import ClusterQuotaDeleteAPIInput
from openapi_client.models.cluster_quota_upsert_api_input import ClusterQuotaUpsertAPIInput
from openapi_client.models.cluster_update_api_input import ClusterUpdateAPIInput
from openapi_client.models.dashboard_chart_create_api_input import DashboardChartCreateAPIInput
from openapi_client.models.dashboard_chart_metric import DashboardChartMetric
from openapi_client.models.dashboard_chart_update_api_input import DashboardChartUpdateAPIInput
from openapi_client.models.dashboard_copy_api_input import DashboardCopyAPIInput
from openapi_client.models.dashboard_create_api_input import DashboardCreateAPIInput
from openapi_client.models.dashboard_dashboard_chart_list_response import DashboardDashboardChartListResponse
from openapi_client.models.dashboard_dashboard_chart_metrics_response import DashboardDashboardChartMetricsResponse
from openapi_client.models.dashboard_dashboard_experiment_field_list_response import DashboardDashboardExperimentFieldListResponse
from openapi_client.models.dashboard_dashboard_experiment_field_update_response import DashboardDashboardExperimentFieldUpdateResponse
from openapi_client.models.dashboard_dashboard_experiment_field_value_list_response import DashboardDashboardExperimentFieldValueListResponse
from openapi_client.models.dashboard_dashboard_experiment_hide_plots_response import DashboardDashboardExperimentHidePlotsResponse
from openapi_client.models.dashboard_dashboard_experiment_list_response import DashboardDashboardExperimentListResponse
from openapi_client.models.dashboard_dashboard_experiment_show_plots_response import DashboardDashboardExperimentShowPlotsResponse
from openapi_client.models.dashboard_dashboard_section_update_response import DashboardDashboardSectionUpdateResponse
from openapi_client.models.dashboard_experiment_add_tags_api_input import DashboardExperimentAddTagsAPIInput
from openapi_client.models.dashboard_experiment_bulk_delete_api_input import DashboardExperimentBulkDeleteAPIInput
from openapi_client.models.dashboard_experiment_bulk_star_api_input import DashboardExperimentBulkStarAPIInput
from openapi_client.models.dashboard_experiment_bulk_terminate_api_input import DashboardExperimentBulkTerminateAPIInput
from openapi_client.models.dashboard_experiment_bulk_unstar_api_input import DashboardExperimentBulkUnstarAPIInput
from openapi_client.models.dashboard_experiment_bulk_update_api_input import DashboardExperimentBulkUpdateAPIInput
from openapi_client.models.dashboard_experiment_field_update_api_input import DashboardExperimentFieldUpdateAPIInput
from openapi_client.models.dashboard_experiment_filter_create_api_input import DashboardExperimentFilterCreateAPIInput
from openapi_client.models.dashboard_experiment_filter_update_api_input import DashboardExperimentFilterUpdateAPIInput
from openapi_client.models.dashboard_experiment_hide_plots_api_input import DashboardExperimentHidePlotsAPIInput
from openapi_client.models.dashboard_experiment_remove_tags_api_input import DashboardExperimentRemoveTagsAPIInput
from openapi_client.models.dashboard_experiment_show_plots_api_input import DashboardExperimentShowPlotsAPIInput
from openapi_client.models.dashboard_experiment_sort_create_api_input import DashboardExperimentSortCreateAPIInput
from openapi_client.models.dashboard_experiment_sort_update_api_input import DashboardExperimentSortUpdateAPIInput
from openapi_client.models.dashboard_experiment_with_metrics import DashboardExperimentWithMetrics
from openapi_client.models.dashboard_field_request import DashboardFieldRequest
from openapi_client.models.dashboard_section_create_api_input import DashboardSectionCreateAPIInput
from openapi_client.models.dashboard_section_update_api_input import DashboardSectionUpdateAPIInput
from openapi_client.models.dashboard_update_api_input import DashboardUpdateAPIInput
from openapi_client.models.dataset_summary_update_api_input import DatasetSummaryUpdateAPIInput
from openapi_client.models.dataset_update_api_input import DatasetUpdateAPIInput
from openapi_client.models.dataset_version_create_api_input import DatasetVersionCreateAPIInput
from openapi_client.models.dataset_version_update_api_input import DatasetVersionUpdateAPIInput
from openapi_client.models.datasetversion_dataset_version_list_response import DatasetversionDatasetVersionListResponse
from openapi_client.models.distributed_experiment_create_api_input import DistributedExperimentCreateAPIInput
from openapi_client.models.distributed_experiment_distributed_experiment_logs_response import DistributedExperimentDistributedExperimentLogsResponse
from openapi_client.models.distributed_experiment_distributed_experiment_plots_response import DistributedExperimentDistributedExperimentPlotsResponse
from openapi_client.models.distributed_experiment_distributed_experiment_system_metrics_response import DistributedExperimentDistributedExperimentSystemMetricsResponse
from openapi_client.models.experiment_add_tag_api_input import ExperimentAddTagAPIInput
from openapi_client.models.experiment_bulk_delete_api_input import ExperimentBulkDeleteAPIInput
from openapi_client.models.experiment_bulk_star_api_input import ExperimentBulkStarAPIInput
from openapi_client.models.experiment_bulk_terminate_api_input import ExperimentBulkTerminateAPIInput
from openapi_client.models.experiment_bulk_unstar_api_input import ExperimentBulkUnstarAPIInput
from openapi_client.models.experiment_create_api_input import ExperimentCreateAPIInput
from openapi_client.models.experiment_experiment_list_response import ExperimentExperimentListResponse
from openapi_client.models.experiment_experiment_logs_response import ExperimentExperimentLogsResponse
from openapi_client.models.experiment_experiment_metrics_update_response import ExperimentExperimentMetricsUpdateResponse
from openapi_client.models.experiment_experiment_parameter_list_response import ExperimentExperimentParameterListResponse
from openapi_client.models.experiment_experiment_plots_response import ExperimentExperimentPlotsResponse
from openapi_client.models.experiment_experiment_status_idle_response import ExperimentExperimentStatusIdleResponse
from openapi_client.models.experiment_experiment_system_metrics_response import ExperimentExperimentSystemMetricsResponse
from openapi_client.models.experiment_metric_entry import ExperimentMetricEntry
from openapi_client.models.experiment_metrics_update_api_input import ExperimentMetricsUpdateAPIInput
from openapi_client.models.experiment_progress_update_api_input import ExperimentProgressUpdateAPIInput
from openapi_client.models.experiment_remove_tag_api_input import ExperimentRemoveTagAPIInput
from openapi_client.models.experiment_update_api_input import ExperimentUpdateAPIInput
from openapi_client.models.experiment_update_hyperparameters_api_input import ExperimentUpdateHyperparametersAPIInput
from openapi_client.models.experiment_yaml_import_api_input import ExperimentYamlImportAPIInput
from openapi_client.models.external_bit_bucket_config_response import ExternalBitBucketConfigResponse
from openapi_client.models.external_git_branch_list_response import ExternalGitBranchListResponse
from openapi_client.models.external_git_commit_list_response import ExternalGitCommitListResponse
from openapi_client.models.external_git_hub_code_refs_response import ExternalGitHubCodeRefsResponse
from openapi_client.models.external_git_hub_config_response import ExternalGitHubConfigResponse
from openapi_client.models.external_git_lab_config_response import ExternalGitLabConfigResponse
from openapi_client.models.external_git_repository_list_response import ExternalGitRepositoryListResponse
from openapi_client.models.external_google_config_response import ExternalGoogleConfigResponse
from openapi_client.models.external_slack_config_response import ExternalSlackConfigResponse
from openapi_client.models.gs_dataset_create_api_input import GSDatasetCreateAPIInput
from openapi_client.models.git_lab_o_auth_authorize_api_input import GitLabOAuthAuthorizeAPIInput
from openapi_client.models.git_lab_token_authorize_api_input import GitLabTokenAuthorizeAPIInput
from openapi_client.models.influxdb_current_system_metric import InfluxdbCurrentSystemMetric
from openapi_client.models.influxdb_experiment_plot_metric import InfluxdbExperimentPlotMetric
from openapi_client.models.influxdb_metric_legacy import InfluxdbMetricLegacy
from openapi_client.models.influxdb_sweep_log import InfluxdbSweepLog
from openapi_client.models.influxdb_system_metric import InfluxdbSystemMetric
from openapi_client.models.influxdb_system_metric_list import InfluxdbSystemMetricList
from openapi_client.models.influxdb_workload_log import InfluxdbWorkloadLog
from openapi_client.models.invitation_token_validate_api_input import InvitationTokenValidateAPIInput
from openapi_client.models.kernel_image_publish_new_managed_image_api_input import KernelImagePublishNewManagedImageAPIInput
from openapi_client.models.kernel_kernel_image_list_response import KernelKernelImageListResponse
from openapi_client.models.kernel_kernel_resource_spec_list_response import KernelKernelResourceSpecListResponse
from openapi_client.models.kernel_resource_spec_create_api_input import KernelResourceSpecCreateAPIInput
from openapi_client.models.kernel_resource_spec_update_api_input import KernelResourceSpecUpdateAPIInput
from openapi_client.models.local_experiment_create_api_input import LocalExperimentCreateAPIInput
from openapi_client.models.local_experiment_finish_api_input import LocalExperimentFinishAPIInput
from openapi_client.models.model_add_tags_api_input import ModelAddTagsAPIInput
from openapi_client.models.model_create_api_input import ModelCreateAPIInput
from openapi_client.models.model_model_add_tag_response import ModelModelAddTagResponse
from openapi_client.models.model_model_remove_tag_response import ModelModelRemoveTagResponse
from openapi_client.models.model_remove_tags_api_input import ModelRemoveTagsAPIInput
from openapi_client.models.model_repository_create_api_input import ModelRepositoryCreateAPIInput
from openapi_client.models.model_repository_update_api_input import ModelRepositoryUpdateAPIInput
from openapi_client.models.model_service_create_api_input import ModelServiceCreateAPIInput
from openapi_client.models.model_service_delete_pod_api_input import ModelServiceDeletePodAPIInput
from openapi_client.models.model_service_gateway_update_api_input import ModelServiceGatewayUpdateAPIInput
from openapi_client.models.model_service_revision_create_api_input import ModelServiceRevisionCreateAPIInput
from openapi_client.models.model_service_revision_terminate_replica_api_input import ModelServiceRevisionTerminateReplicaAPIInput
from openapi_client.models.model_service_revision_update_api_input import ModelServiceRevisionUpdateAPIInput
from openapi_client.models.model_service_rollout_create_api_input import ModelServiceRolloutCreateAPIInput
from openapi_client.models.model_service_rollout_update_status_api_input import ModelServiceRolloutUpdateStatusAPIInput
from openapi_client.models.model_service_update_api_input import ModelServiceUpdateAPIInput
from openapi_client.models.model_update_api_input import ModelUpdateAPIInput
from openapi_client.models.modelservice_model_service_list_response import ModelserviceModelServiceListResponse
from openapi_client.models.modelservice_model_service_logs_response import ModelserviceModelServiceLogsResponse
from openapi_client.models.modelservice_model_service_revision_list_response import ModelserviceModelServiceRevisionListResponse
from openapi_client.models.modelservice_model_service_serving_metrics_response import ModelserviceModelServiceServingMetricsResponse
from openapi_client.models.modelservice_model_service_system_metrics_response import ModelserviceModelServiceSystemMetricsResponse
from openapi_client.models.modelservice_model_service_workload_list_response import ModelserviceModelServiceWorkloadListResponse
from openapi_client.models.on_premise_dataset_create_api_input import OnPremiseDatasetCreateAPIInput
from openapi_client.models.organization_aws_credentials_add_api_input import OrganizationAWSCredentialsAddAPIInput
from openapi_client.models.organization_artifactory_credentials_add_api_input import OrganizationArtifactoryCredentialsAddAPIInput
from openapi_client.models.organization_create_api_input import OrganizationCreateAPIInput
from openapi_client.models.organization_docker_credentials_add_api_input import OrganizationDockerCredentialsAddAPIInput
from openapi_client.models.organization_git_hub_credentials_add_api_input import OrganizationGitHubCredentialsAddAPIInput
from openapi_client.models.organization_history_add_api_input import OrganizationHistoryAddAPIInput
from openapi_client.models.organization_member_add_api_input import OrganizationMemberAddAPIInput
from openapi_client.models.organization_member_bulk_add_api200_response import OrganizationMemberBulkAddAPI200Response
from openapi_client.models.organization_member_bulk_add_api_input import OrganizationMemberBulkAddAPIInput
from openapi_client.models.organization_member_update_permission_api_input import OrganizationMemberUpdatePermissionAPIInput
from openapi_client.models.organization_organization_billing_cluster_list_response import OrganizationOrganizationBillingClusterListResponse
from openapi_client.models.organization_organization_billing_history_list_response import OrganizationOrganizationBillingHistoryListResponse
from openapi_client.models.organization_organization_billing_member_list_response import OrganizationOrganizationBillingMemberListResponse
from openapi_client.models.organization_organization_billing_past_due_response import OrganizationOrganizationBillingPastDueResponse
from openapi_client.models.organization_organization_billing_read_response import OrganizationOrganizationBillingReadResponse
from openapi_client.models.organization_organization_list_response import OrganizationOrganizationListResponse
from openapi_client.models.organization_organization_me_response import OrganizationOrganizationMeResponse
from openapi_client.models.organization_organization_member_list_response import OrganizationOrganizationMemberListResponse
from openapi_client.models.organization_organization_member_update_permission_response import OrganizationOrganizationMemberUpdatePermissionResponse
from openapi_client.models.organization_plan_update_api_input import OrganizationPlanUpdateAPIInput
from openapi_client.models.organization_slack_credentials_add_api_input import OrganizationSlackCredentialsAddAPIInput
from openapi_client.models.organization_update_api_input import OrganizationUpdateAPIInput
from openapi_client.models.orm_access_control_config import OrmAccessControlConfig
from openapi_client.models.orm_access_control_policy import OrmAccessControlPolicy
from openapi_client.models.orm_access_control_policy_edges import OrmAccessControlPolicyEdges
from openapi_client.models.orm_access_token import OrmAccessToken
from openapi_client.models.orm_access_token_edges import OrmAccessTokenEdges
from openapi_client.models.orm_autoscaler_config import OrmAutoscalerConfig
from openapi_client.models.orm_billing_history import OrmBillingHistory
from openapi_client.models.orm_billing_history_edges import OrmBillingHistoryEdges
from openapi_client.models.orm_cluster_quota import OrmClusterQuota
from openapi_client.models.orm_cluster_quota_edges import OrmClusterQuotaEdges
from openapi_client.models.orm_credit_earn_history import OrmCreditEarnHistory
from openapi_client.models.orm_credit_earn_history_edges import OrmCreditEarnHistoryEdges
from openapi_client.models.orm_dataset import OrmDataset
from openapi_client.models.orm_dataset_edges import OrmDatasetEdges
from openapi_client.models.orm_dataset_summary import OrmDatasetSummary
from openapi_client.models.orm_dataset_summary_edges import OrmDatasetSummaryEdges
from openapi_client.models.orm_distributed_py_torch_spec import OrmDistributedPyTorchSpec
from openapi_client.models.orm_distributed_spec import OrmDistributedSpec
from openapi_client.models.orm_duration_value import OrmDurationValue
from openapi_client.models.orm_early_stopping_setting import OrmEarlyStoppingSetting
from openapi_client.models.orm_early_stopping_spec import OrmEarlyStoppingSpec
from openapi_client.models.orm_env_var import OrmEnvVar
from openapi_client.models.orm_env_vars import OrmEnvVars
from openapi_client.models.orm_execution_environment import OrmExecutionEnvironment
from openapi_client.models.orm_experiment import OrmExperiment
from openapi_client.models.orm_experiment_edges import OrmExperimentEdges
from openapi_client.models.orm_experiment_filter_values import OrmExperimentFilterValues
from openapi_client.models.orm_experiment_metrics_summary import OrmExperimentMetricsSummary
from openapi_client.models.orm_group import OrmGroup
from openapi_client.models.orm_group_edges import OrmGroupEdges
from openapi_client.models.orm_hyperparameter import OrmHyperparameter
from openapi_client.models.orm_hyperparameters import OrmHyperparameters
from openapi_client.models.orm_kernel_cluster import OrmKernelCluster
from openapi_client.models.orm_kernel_cluster_config import OrmKernelClusterConfig
from openapi_client.models.orm_kernel_cluster_config_ingress import OrmKernelClusterConfigIngress
from openapi_client.models.orm_kernel_cluster_config_nodes import OrmKernelClusterConfigNodes
from openapi_client.models.orm_kernel_cluster_config_service import OrmKernelClusterConfigService
from openapi_client.models.orm_kernel_cluster_config_storage_class import OrmKernelClusterConfigStorageClass
from openapi_client.models.orm_kernel_cluster_edges import OrmKernelClusterEdges
from openapi_client.models.orm_kernel_cluster_node import OrmKernelClusterNode
from openapi_client.models.orm_kernel_cluster_node_edges import OrmKernelClusterNodeEdges
from openapi_client.models.orm_kernel_cluster_select_policies import OrmKernelClusterSelectPolicies
from openapi_client.models.orm_kernel_cluster_select_policy import OrmKernelClusterSelectPolicy
from openapi_client.models.orm_kernel_cluster_storage import OrmKernelClusterStorage
from openapi_client.models.orm_kernel_cluster_storage_edges import OrmKernelClusterStorageEdges
from openapi_client.models.orm_kernel_image import OrmKernelImage
from openapi_client.models.orm_kernel_image_edges import OrmKernelImageEdges
from openapi_client.models.orm_kernel_resource_spec import OrmKernelResourceSpec
from openapi_client.models.orm_kernel_resource_spec_edges import OrmKernelResourceSpecEdges
from openapi_client.models.orm_kernel_resource_spec_field import OrmKernelResourceSpecField
from openapi_client.models.orm_kernel_resource_spec_node_selector import OrmKernelResourceSpecNodeSelector
from openapi_client.models.orm_key_metric import OrmKeyMetric
from openapi_client.models.orm_key_metrics import OrmKeyMetrics
from openapi_client.models.orm_local_execution_spec import OrmLocalExecutionSpec
from openapi_client.models.orm_model import OrmModel
from openapi_client.models.orm_model_edges import OrmModelEdges
from openapi_client.models.orm_model_generated_experiment import OrmModelGeneratedExperiment
from openapi_client.models.orm_model_generated_experiment_edges import OrmModelGeneratedExperimentEdges
from openapi_client.models.orm_model_repository import OrmModelRepository
from openapi_client.models.orm_model_repository_edges import OrmModelRepositoryEdges
from openapi_client.models.orm_model_service import OrmModelService
from openapi_client.models.orm_model_service_edges import OrmModelServiceEdges
from openapi_client.models.orm_model_service_gateway import OrmModelServiceGateway
from openapi_client.models.orm_model_service_gateway_edges import OrmModelServiceGatewayEdges
from openapi_client.models.orm_model_service_gateway_traffic_split_entry import OrmModelServiceGatewayTrafficSplitEntry
from openapi_client.models.orm_model_service_revision import OrmModelServiceRevision
from openapi_client.models.orm_model_service_revision_deployment_spec import OrmModelServiceRevisionDeploymentSpec
from openapi_client.models.orm_model_service_revision_deployment_type_custom_image import OrmModelServiceRevisionDeploymentTypeCustomImage
from openapi_client.models.orm_model_service_revision_deployment_type_vessl_model import OrmModelServiceRevisionDeploymentTypeVESSLModel
from openapi_client.models.orm_model_service_revision_edges import OrmModelServiceRevisionEdges
from openapi_client.models.orm_model_service_rollout import OrmModelServiceRollout
from openapi_client.models.orm_model_service_rollout_edges import OrmModelServiceRolloutEdges
from openapi_client.models.orm_model_service_rollout_revision_reference import OrmModelServiceRolloutRevisionReference
from openapi_client.models.orm_model_service_rollout_spec import OrmModelServiceRolloutSpec
from openapi_client.models.orm_model_service_rollout_step_create_new_revision import OrmModelServiceRolloutStepCreateNewRevision
from openapi_client.models.orm_model_service_rollout_step_send_notification import OrmModelServiceRolloutStepSendNotification
from openapi_client.models.orm_model_service_rollout_step_status import OrmModelServiceRolloutStepStatus
from openapi_client.models.orm_model_service_rollout_step_update_endpoint import OrmModelServiceRolloutStepUpdateEndpoint
from openapi_client.models.orm_model_service_rollout_step_update_endpoint_revision_entry import OrmModelServiceRolloutStepUpdateEndpointRevisionEntry
from openapi_client.models.orm_model_service_rollout_step_update_revisions import OrmModelServiceRolloutStepUpdateRevisions
from openapi_client.models.orm_model_service_rollout_step_update_revisions_revision_target import OrmModelServiceRolloutStepUpdateRevisionsRevisionTarget
from openapi_client.models.orm_model_service_rollout_step_wait import OrmModelServiceRolloutStepWait
from openapi_client.models.orm_model_service_rollout_step_wrapper import OrmModelServiceRolloutStepWrapper
from openapi_client.models.orm_objective_step_median import OrmObjectiveStepMedian
from openapi_client.models.orm_on_premise_volume_config import OrmOnPremiseVolumeConfig
from openapi_client.models.orm_on_premise_volume_config_flex_volume import OrmOnPremiseVolumeConfigFlexVolume
from openapi_client.models.orm_on_premise_volume_config_flex_volume_options_inner import OrmOnPremiseVolumeConfigFlexVolumeOptionsInner
from openapi_client.models.orm_on_premise_volume_config_google_disk import OrmOnPremiseVolumeConfigGoogleDisk
from openapi_client.models.orm_on_premise_volume_config_host_path import OrmOnPremiseVolumeConfigHostPath
from openapi_client.models.orm_on_premise_volume_config_nfs import OrmOnPremiseVolumeConfigNFS
from openapi_client.models.orm_organization import OrmOrganization
from openapi_client.models.orm_organization_credentials import OrmOrganizationCredentials
from openapi_client.models.orm_organization_credentials_edges import OrmOrganizationCredentialsEdges
from openapi_client.models.orm_organization_edges import OrmOrganizationEdges
from openapi_client.models.orm_organization_kernel_cluster import OrmOrganizationKernelCluster
from openapi_client.models.orm_organization_kernel_cluster_edges import OrmOrganizationKernelClusterEdges
from openapi_client.models.orm_page_info_with_count import OrmPageInfoWithCount
from openapi_client.models.orm_parameter import OrmParameter
from openapi_client.models.orm_pipeline import OrmPipeline
from openapi_client.models.orm_pipeline_edges import OrmPipelineEdges
from openapi_client.models.orm_pipeline_execution import OrmPipelineExecution
from openapi_client.models.orm_pipeline_execution_edges import OrmPipelineExecutionEdges
from openapi_client.models.orm_pipeline_spec import OrmPipelineSpec
from openapi_client.models.orm_pipeline_spec_edges import OrmPipelineSpecEdges
from openapi_client.models.orm_pipeline_step_execution import OrmPipelineStepExecution
from openapi_client.models.orm_pipeline_step_execution_edges import OrmPipelineStepExecutionEdges
from openapi_client.models.orm_pipeline_step_external_service import OrmPipelineStepExternalService
from openapi_client.models.orm_pipeline_step_external_service_edges import OrmPipelineStepExternalServiceEdges
from openapi_client.models.orm_pipeline_step_if import OrmPipelineStepIf
from openapi_client.models.orm_pipeline_step_if_condition import OrmPipelineStepIfCondition
from openapi_client.models.orm_pipeline_step_if_edges import OrmPipelineStepIfEdges
from openapi_client.models.orm_pipeline_step_if_variable import OrmPipelineStepIfVariable
from openapi_client.models.orm_pipeline_step_jupyter_visualization import OrmPipelineStepJupyterVisualization
from openapi_client.models.orm_pipeline_step_jupyter_visualization_edges import OrmPipelineStepJupyterVisualizationEdges
from openapi_client.models.orm_pipeline_step_manual_input import OrmPipelineStepManualInput
from openapi_client.models.orm_pipeline_step_manual_input_edges import OrmPipelineStepManualInputEdges
from openapi_client.models.orm_pipeline_step_manual_judgment import OrmPipelineStepManualJudgment
from openapi_client.models.orm_pipeline_step_manual_judgment_edges import OrmPipelineStepManualJudgmentEdges
from openapi_client.models.orm_pipeline_step_run import OrmPipelineStepRun
from openapi_client.models.orm_pipeline_step_run_edges import OrmPipelineStepRunEdges
from openapi_client.models.orm_pipeline_step_spec import OrmPipelineStepSpec
from openapi_client.models.orm_pipeline_step_spec_edges import OrmPipelineStepSpecEdges
from openapi_client.models.orm_pipeline_step_type import OrmPipelineStepType
from openapi_client.models.orm_pipeline_step_type_edges import OrmPipelineStepTypeEdges
from openapi_client.models.orm_pipeline_trigger import OrmPipelineTrigger
from openapi_client.models.orm_pipeline_trigger_cron_spec import OrmPipelineTriggerCronSpec
from openapi_client.models.orm_pipeline_trigger_edges import OrmPipelineTriggerEdges
from openapi_client.models.orm_pricing_plan import OrmPricingPlan
from openapi_client.models.orm_pricing_plan_edges import OrmPricingPlanEdges
from openapi_client.models.orm_project import OrmProject
from openapi_client.models.orm_project_dashboard import OrmProjectDashboard
from openapi_client.models.orm_project_dashboard_chart import OrmProjectDashboardChart
from openapi_client.models.orm_project_dashboard_chart_edges import OrmProjectDashboardChartEdges
from openapi_client.models.orm_project_dashboard_chart_section import OrmProjectDashboardChartSection
from openapi_client.models.orm_project_dashboard_chart_section_edges import OrmProjectDashboardChartSectionEdges
from openapi_client.models.orm_project_dashboard_edges import OrmProjectDashboardEdges
from openapi_client.models.orm_project_dashboard_experiment_field_config import OrmProjectDashboardExperimentFieldConfig
from openapi_client.models.orm_project_dashboard_experiment_field_config_edges import OrmProjectDashboardExperimentFieldConfigEdges
from openapi_client.models.orm_project_dashboard_experiment_filter import OrmProjectDashboardExperimentFilter
from openapi_client.models.orm_project_dashboard_experiment_filter_edges import OrmProjectDashboardExperimentFilterEdges
from openapi_client.models.orm_project_dashboard_experiment_sort import OrmProjectDashboardExperimentSort
from openapi_client.models.orm_project_dashboard_experiment_sort_edges import OrmProjectDashboardExperimentSortEdges
from openapi_client.models.orm_project_dataset import OrmProjectDataset
from openapi_client.models.orm_project_dataset_edges import OrmProjectDatasetEdges
from openapi_client.models.orm_project_edges import OrmProjectEdges
from openapi_client.models.orm_project_experiment_field import OrmProjectExperimentField
from openapi_client.models.orm_project_experiment_field_edges import OrmProjectExperimentFieldEdges
from openapi_client.models.orm_project_repository import OrmProjectRepository
from openapi_client.models.orm_project_repository_edges import OrmProjectRepositoryEdges
from openapi_client.models.orm_range import OrmRange
from openapi_client.models.orm_region import OrmRegion
from openapi_client.models.orm_region_edges import OrmRegionEdges
from openapi_client.models.orm_reset_password_token import OrmResetPasswordToken
from openapi_client.models.orm_reset_password_token_edges import OrmResetPasswordTokenEdges
from openapi_client.models.orm_scheduled_pipeline_execution import OrmScheduledPipelineExecution
from openapi_client.models.orm_scheduled_pipeline_execution_edges import OrmScheduledPipelineExecutionEdges
from openapi_client.models.orm_service import OrmService
from openapi_client.models.orm_service_edges import OrmServiceEdges
from openapi_client.models.orm_step_median import OrmStepMedian
from openapi_client.models.orm_storage import OrmStorage
from openapi_client.models.orm_storage_edges import OrmStorageEdges
from openapi_client.models.orm_stripe_billing_history import OrmStripeBillingHistory
from openapi_client.models.orm_stripe_billing_history_edges import OrmStripeBillingHistoryEdges
from openapi_client.models.orm_subject import OrmSubject
from openapi_client.models.orm_subject_edges import OrmSubjectEdges
from openapi_client.models.orm_suggestion_histories import OrmSuggestionHistories
from openapi_client.models.orm_suggestion_history import OrmSuggestionHistory
from openapi_client.models.orm_suggestion_history_parameter import OrmSuggestionHistoryParameter
from openapi_client.models.orm_sweep import OrmSweep
from openapi_client.models.orm_sweep_edges import OrmSweepEdges
from openapi_client.models.orm_sweep_history import OrmSweepHistory
from openapi_client.models.orm_sweep_history_edges import OrmSweepHistoryEdges
from openapi_client.models.orm_sweep_objective import OrmSweepObjective
from openapi_client.models.orm_sweep_search_space import OrmSweepSearchSpace
from openapi_client.models.orm_tag import OrmTag
from openapi_client.models.orm_tag_edges import OrmTagEdges
from openapi_client.models.orm_tag_group import OrmTagGroup
from openapi_client.models.orm_tag_group_edges import OrmTagGroupEdges
from openapi_client.models.orm_user import OrmUser
from openapi_client.models.orm_user_edges import OrmUserEdges
from openapi_client.models.orm_user_group import OrmUserGroup
from openapi_client.models.orm_user_group_edges import OrmUserGroupEdges
from openapi_client.models.orm_user_history import OrmUserHistory
from openapi_client.models.orm_user_history_edges import OrmUserHistoryEdges
from openapi_client.models.orm_user_organization import OrmUserOrganization
from openapi_client.models.orm_user_organization_edges import OrmUserOrganizationEdges
from openapi_client.models.orm_volume import OrmVolume
from openapi_client.models.orm_volume_claim import OrmVolumeClaim
from openapi_client.models.orm_volume_claim_edges import OrmVolumeClaimEdges
from openapi_client.models.orm_volume_edges import OrmVolumeEdges
from openapi_client.models.orm_volume_mount_request import OrmVolumeMountRequest
from openapi_client.models.orm_volume_mount_request_source_archive_file import OrmVolumeMountRequestSourceArchiveFile
from openapi_client.models.orm_volume_mount_request_source_code import OrmVolumeMountRequestSourceCode
from openapi_client.models.orm_volume_mount_request_source_dataset import OrmVolumeMountRequestSourceDataset
from openapi_client.models.orm_volume_mount_request_source_dataset_version import OrmVolumeMountRequestSourceDatasetVersion
from openapi_client.models.orm_volume_mount_request_source_model_volume import OrmVolumeMountRequestSourceModelVolume
from openapi_client.models.orm_volume_mount_request_source_object_storage import OrmVolumeMountRequestSourceObjectStorage
from openapi_client.models.orm_volume_mount_request_source_volume import OrmVolumeMountRequestSourceVolume
from openapi_client.models.orm_volume_mount_requests import OrmVolumeMountRequests
from openapi_client.models.orm_volume_source_archive_file import OrmVolumeSourceArchiveFile
from openapi_client.models.orm_volume_source_code import OrmVolumeSourceCode
from openapi_client.models.orm_volume_source_dataset import OrmVolumeSourceDataset
from openapi_client.models.orm_volume_source_dataset_version import OrmVolumeSourceDatasetVersion
from openapi_client.models.orm_volume_source_model_volume import OrmVolumeSourceModelVolume
from openapi_client.models.orm_volume_source_object_storage import OrmVolumeSourceObjectStorage
from openapi_client.models.orm_volume_source_volume import OrmVolumeSourceVolume
from openapi_client.models.orm_withdraw_history_organization import OrmWithdrawHistoryOrganization
from openapi_client.models.orm_withdraw_history_organization_edges import OrmWithdrawHistoryOrganizationEdges
from openapi_client.models.orm_withdraw_history_workload import OrmWithdrawHistoryWorkload
from openapi_client.models.orm_withdraw_history_workload_edges import OrmWithdrawHistoryWorkloadEdges
from openapi_client.models.orm_workload import OrmWorkload
from openapi_client.models.orm_workload_edges import OrmWorkloadEdges
from openapi_client.models.orm_workload_endpoint import OrmWorkloadEndpoint
from openapi_client.models.orm_workload_history import OrmWorkloadHistory
from openapi_client.models.orm_workload_history_edges import OrmWorkloadHistoryEdges
from openapi_client.models.orm_workload_pod_info import OrmWorkloadPodInfo
from openapi_client.models.orm_workload_port import OrmWorkloadPort
from openapi_client.models.orm_workload_ports import OrmWorkloadPorts
from openapi_client.models.orm_workloads_summary import OrmWorkloadsSummary
from openapi_client.models.orm_workspace import OrmWorkspace
from openapi_client.models.orm_workspace_backup import OrmWorkspaceBackup
from openapi_client.models.orm_workspace_backup_edges import OrmWorkspaceBackupEdges
from openapi_client.models.orm_workspace_edges import OrmWorkspaceEdges
from openapi_client.models.orm_workspace_port import OrmWorkspacePort
from openapi_client.models.orm_workspace_ports import OrmWorkspacePorts
from openapi_client.models.pipeline_context import PipelineContext
from openapi_client.models.pipeline_context_metadata import PipelineContextMetadata
from openapi_client.models.pipeline_create_api_input import PipelineCreateAPIInput
from openapi_client.models.pipeline_external_service_spec import PipelineExternalServiceSpec
from openapi_client.models.pipeline_if_spec import PipelineIfSpec
from openapi_client.models.pipeline_jupyter_visualization_spec import PipelineJupyterVisualizationSpec
from openapi_client.models.pipeline_manual_input_spec import PipelineManualInputSpec
from openapi_client.models.pipeline_pipeline_execution_list_response import PipelinePipelineExecutionListResponse
from openapi_client.models.pipeline_pipeline_list_response import PipelinePipelineListResponse
from openapi_client.models.pipeline_pipeline_single_variable_spec import PipelinePipelineSingleVariableSpec
from openapi_client.models.pipeline_pipeline_spec_create_staged_revision_response import PipelinePipelineSpecCreateStagedRevisionResponse
from openapi_client.models.pipeline_pipeline_spec_update_staged_revision_response import PipelinePipelineSpecUpdateStagedRevisionResponse
from openapi_client.models.pipeline_pipeline_step_type_list_response import PipelinePipelineStepTypeListResponse
from openapi_client.models.pipeline_run_spec import PipelineRunSpec
from openapi_client.models.pipeline_spec_publish_revision_api_input import PipelineSpecPublishRevisionAPIInput
from openapi_client.models.pipeline_spec_update_staged_revision_api_input import PipelineSpecUpdateStagedRevisionAPIInput
from openapi_client.models.pipeline_step_create_api_input import PipelineStepCreateAPIInput
from openapi_client.models.pipeline_step_dependency_add_api_input import PipelineStepDependencyAddAPIInput
from openapi_client.models.pipeline_step_dependency_remove_api_input import PipelineStepDependencyRemoveAPIInput
from openapi_client.models.pipeline_step_execution_variable_update_api_input import PipelineStepExecutionVariableUpdateAPIInput
from openapi_client.models.pipeline_step_unmount_volume_claim_api_input import PipelineStepUnmountVolumeClaimAPIInput
from openapi_client.models.pipeline_step_update_api_input import PipelineStepUpdateAPIInput
from openapi_client.models.pipeline_trigger_create_api_input import PipelineTriggerCreateAPIInput
from openapi_client.models.pipeline_trigger_dispatch_api_input import PipelineTriggerDispatchAPIInput
from openapi_client.models.pipeline_trigger_toggle_enabled_api_input import PipelineTriggerToggleEnabledAPIInput
from openapi_client.models.pipeline_update_api_input import PipelineUpdateAPIInput
from openapi_client.models.project_add_project_dataset_api_input import ProjectAddProjectDatasetAPIInput
from openapi_client.models.project_add_project_repository_api_input import ProjectAddProjectRepositoryAPIInput
from openapi_client.models.project_create_api_input import ProjectCreateAPIInput
from openapi_client.models.project_project_branch_list_response import ProjectProjectBranchListResponse
from openapi_client.models.project_project_commit_list_response import ProjectProjectCommitListResponse
from openapi_client.models.project_project_repository_branch_list_response import ProjectProjectRepositoryBranchListResponse
from openapi_client.models.project_project_repository_commit_list_response import ProjectProjectRepositoryCommitListResponse
from openapi_client.models.project_remove_project_dataset_api_input import ProjectRemoveProjectDatasetAPIInput
from openapi_client.models.project_remove_project_repository_api_input import ProjectRemoveProjectRepositoryAPIInput
from openapi_client.models.project_update_api_input import ProjectUpdateAPIInput
from openapi_client.models.prometheusquery_cluster_metric_sample import PrometheusqueryClusterMetricSample
from openapi_client.models.prometheusquery_cluster_metrics import PrometheusqueryClusterMetrics
from openapi_client.models.prometheusquery_cluster_metrics_summary import PrometheusqueryClusterMetricsSummary
from openapi_client.models.prometheusquery_latest_node_metric_sample import PrometheusqueryLatestNodeMetricSample
from openapi_client.models.prometheusquery_node_metric_series import PrometheusqueryNodeMetricSeries
from openapi_client.models.prometheusquery_node_metric_series_group import PrometheusqueryNodeMetricSeriesGroup
from openapi_client.models.prometheusquery_node_metrics import PrometheusqueryNodeMetrics
from openapi_client.models.prometheusquery_node_resource_metric_sample import PrometheusqueryNodeResourceMetricSample
from openapi_client.models.prometheusquery_resource_metric_sample import PrometheusqueryResourceMetricSample
from openapi_client.models.prometheusquery_sample import PrometheusquerySample
from openapi_client.models.prometheusquery_workload_sample_series import PrometheusqueryWorkloadSampleSeries
from openapi_client.models.proto_branch import ProtoBranch
from openapi_client.models.proto_commit import ProtoCommit
from openapi_client.models.proto_commit_author import ProtoCommitAuthor
from openapi_client.models.proto_project_dataset import ProtoProjectDataset
from openapi_client.models.proto_project_repository_request import ProtoProjectRepositoryRequest
from openapi_client.models.proto_repository import ProtoRepository
from openapi_client.models.proto_tag import ProtoTag
from openapi_client.models.redis_entity_info import RedisEntityInfo
from openapi_client.models.redis_organization_activity import RedisOrganizationActivity
from openapi_client.models.redis_organization_history import RedisOrganizationHistory
from openapi_client.models.redis_organization_history_member import RedisOrganizationHistoryMember
from openapi_client.models.redis_project_key_metric import RedisProjectKeyMetric
from openapi_client.models.resend_verify_email_api_input import ResendVerifyEmailAPIInput
from openapi_client.models.reset_password_token_redeem_api_input import ResetPasswordTokenRedeemAPIInput
from openapi_client.models.response_billing_history_info import ResponseBillingHistoryInfo
from openapi_client.models.response_cluster_quota import ResponseClusterQuota
from openapi_client.models.response_code_repository import ResponseCodeRepository
from openapi_client.models.response_dashboard_basic import ResponseDashboardBasic
from openapi_client.models.response_dashboard_chart import ResponseDashboardChart
from openapi_client.models.response_dashboard_chart_field import ResponseDashboardChartField
from openapi_client.models.response_dashboard_chart_section import ResponseDashboardChartSection
from openapi_client.models.response_dashboard_detail import ResponseDashboardDetail
from openapi_client.models.response_dashboard_experiment_field import ResponseDashboardExperimentField
from openapi_client.models.response_dashboard_experiment_filter_response import ResponseDashboardExperimentFilterResponse
from openapi_client.models.response_dashboard_experiment_filter_values import ResponseDashboardExperimentFilterValues
from openapi_client.models.response_dashboard_experiment_sort_response import ResponseDashboardExperimentSortResponse
from openapi_client.models.response_dashboard_list import ResponseDashboardList
from openapi_client.models.response_dashboard_status import ResponseDashboardStatus
from openapi_client.models.response_dataset_info import ResponseDatasetInfo
from openapi_client.models.response_dataset_info_detail import ResponseDatasetInfoDetail
from openapi_client.models.response_dataset_info_list import ResponseDatasetInfoList
from openapi_client.models.response_dataset_source import ResponseDatasetSource
from openapi_client.models.response_dataset_summary import ResponseDatasetSummary
from openapi_client.models.response_dataset_version_info import ResponseDatasetVersionInfo
from openapi_client.models.response_experiment_info import ResponseExperimentInfo
from openapi_client.models.response_experiment_list_response import ResponseExperimentListResponse
from openapi_client.models.response_experiment_plot_file import ResponseExperimentPlotFile
from openapi_client.models.response_field_object_value import ResponseFieldObjectValue
from openapi_client.models.response_git_hub_code_ref import ResponseGitHubCodeRef
from openapi_client.models.response_kernel_cluster import ResponseKernelCluster
from openapi_client.models.response_kernel_cluster_info import ResponseKernelClusterInfo
from openapi_client.models.response_kernel_cluster_node import ResponseKernelClusterNode
from openapi_client.models.response_kernel_cluster_node_info import ResponseKernelClusterNodeInfo
from openapi_client.models.response_kernel_cluster_node_info_v2 import ResponseKernelClusterNodeInfoV2
from openapi_client.models.response_kernel_cluster_system_metrics import ResponseKernelClusterSystemMetrics
from openapi_client.models.response_kernel_cluster_usage_info import ResponseKernelClusterUsageInfo
from openapi_client.models.response_kernel_image import ResponseKernelImage
from openapi_client.models.response_kernel_resource_spec import ResponseKernelResourceSpec
from openapi_client.models.response_kubernetes_resource_info import ResponseKubernetesResourceInfo
from openapi_client.models.response_model_detail import ResponseModelDetail
from openapi_client.models.response_model_info import ResponseModelInfo
from openapi_client.models.response_model_info_list import ResponseModelInfoList
from openapi_client.models.response_model_repository_detail import ResponseModelRepositoryDetail
from openapi_client.models.response_model_repository_list import ResponseModelRepositoryList
from openapi_client.models.response_model_service_gateway_info import ResponseModelServiceGatewayInfo
from openapi_client.models.response_model_service_info import ResponseModelServiceInfo
from openapi_client.models.response_model_service_revision import ResponseModelServiceRevision
from openapi_client.models.response_model_service_rollout_info import ResponseModelServiceRolloutInfo
from openapi_client.models.response_model_service_rollout_list import ResponseModelServiceRolloutList
from openapi_client.models.response_model_service_rollout_list_item import ResponseModelServiceRolloutListItem
from openapi_client.models.response_model_service_rollout_related_revision_info import ResponseModelServiceRolloutRelatedRevisionInfo
from openapi_client.models.response_model_summary import ResponseModelSummary
from openapi_client.models.response_my_user import ResponseMyUser
from openapi_client.models.response_my_workload import ResponseMyWorkload
from openapi_client.models.response_on_premise_kernel_cluster_info import ResponseOnPremiseKernelClusterInfo
from openapi_client.models.response_organization import ResponseOrganization
from openapi_client.models.response_organization_activities import ResponseOrganizationActivities
from openapi_client.models.response_organization_credentials_info import ResponseOrganizationCredentialsInfo
from openapi_client.models.response_organization_credentials_info_list import ResponseOrganizationCredentialsInfoList
from openapi_client.models.response_organization_history import ResponseOrganizationHistory
from openapi_client.models.response_organization_info import ResponseOrganizationInfo
from openapi_client.models.response_organization_kernel_cluster_info import ResponseOrganizationKernelClusterInfo
from openapi_client.models.response_organization_member import ResponseOrganizationMember
from openapi_client.models.response_parameter_info import ResponseParameterInfo
from openapi_client.models.response_pipeline import ResponsePipeline
from openapi_client.models.response_pipeline_execution import ResponsePipelineExecution
from openapi_client.models.response_pipeline_spec import ResponsePipelineSpec
from openapi_client.models.response_pipeline_step import ResponsePipelineStep
from openapi_client.models.response_pipeline_step_dependency import ResponsePipelineStepDependency
from openapi_client.models.response_pipeline_step_execution import ResponsePipelineStepExecution
from openapi_client.models.response_pipeline_step_external_service_result import ResponsePipelineStepExternalServiceResult
from openapi_client.models.response_pipeline_step_external_service_spec import ResponsePipelineStepExternalServiceSpec
from openapi_client.models.response_pipeline_step_if_result import ResponsePipelineStepIfResult
from openapi_client.models.response_pipeline_step_if_spec import ResponsePipelineStepIfSpec
from openapi_client.models.response_pipeline_step_jupyter_visualization_result import ResponsePipelineStepJupyterVisualizationResult
from openapi_client.models.response_pipeline_step_jupyter_visualization_spec import ResponsePipelineStepJupyterVisualizationSpec
from openapi_client.models.response_pipeline_step_manual_input_result import ResponsePipelineStepManualInputResult
from openapi_client.models.response_pipeline_step_manual_input_spec import ResponsePipelineStepManualInputSpec
from openapi_client.models.response_pipeline_step_manual_judgment_result import ResponsePipelineStepManualJudgmentResult
from openapi_client.models.response_pipeline_step_run_result import ResponsePipelineStepRunResult
from openapi_client.models.response_pipeline_step_run_spec import ResponsePipelineStepRunSpec
from openapi_client.models.response_pipeline_step_type import ResponsePipelineStepType
from openapi_client.models.response_pipeline_trigger import ResponsePipelineTrigger
from openapi_client.models.response_pipeline_trigger_cron import ResponsePipelineTriggerCron
from openapi_client.models.response_pipeline_variable import ResponsePipelineVariable
from openapi_client.models.response_pricing_plan import ResponsePricingPlan
from openapi_client.models.response_project import ResponseProject
from openapi_client.models.response_project_dataset import ResponseProjectDataset
from openapi_client.models.response_project_experiment_summary import ResponseProjectExperimentSummary
from openapi_client.models.response_project_info import ResponseProjectInfo
from openapi_client.models.response_project_key_metrics import ResponseProjectKeyMetrics
from openapi_client.models.response_project_list_response import ResponseProjectListResponse
from openapi_client.models.response_project_repository import ResponseProjectRepository
from openapi_client.models.response_reduced_pipeline import ResponseReducedPipeline
from openapi_client.models.response_reduced_pipeline_execution import ResponseReducedPipelineExecution
from openapi_client.models.response_reduced_pipeline_step import ResponseReducedPipelineStep
from openapi_client.models.response_reduced_pipeline_step_execution import ResponseReducedPipelineStepExecution
from openapi_client.models.response_reduced_pipeline_step_jupyter_visualization_spec import ResponseReducedPipelineStepJupyterVisualizationSpec
from openapi_client.models.response_reduced_pipeline_step_run_spec import ResponseReducedPipelineStepRunSpec
from openapi_client.models.response_ssh_key_info import ResponseSSHKeyInfo
from openapi_client.models.response_service_info import ResponseServiceInfo
from openapi_client.models.response_simple_experiment import ResponseSimpleExperiment
from openapi_client.models.response_simple_kernel_cluster_node import ResponseSimpleKernelClusterNode
from openapi_client.models.response_simple_model_info import ResponseSimpleModelInfo
from openapi_client.models.response_simple_model_service_revision import ResponseSimpleModelServiceRevision
from openapi_client.models.response_simple_project import ResponseSimpleProject
from openapi_client.models.response_simple_service_info import ResponseSimpleServiceInfo
from openapi_client.models.response_simple_sweep import ResponseSimpleSweep
from openapi_client.models.response_simple_user import ResponseSimpleUser
from openapi_client.models.response_simple_workspace import ResponseSimpleWorkspace
from openapi_client.models.response_sweep_experiment_info import ResponseSweepExperimentInfo
from openapi_client.models.response_sweep_experiment_summary import ResponseSweepExperimentSummary
from openapi_client.models.response_sweep_history_info import ResponseSweepHistoryInfo
from openapi_client.models.response_sweep_info import ResponseSweepInfo
from openapi_client.models.response_sweep_list_response import ResponseSweepListResponse
from openapi_client.models.response_tag_response import ResponseTagResponse
from openapi_client.models.response_tutorial_response import ResponseTutorialResponse
from openapi_client.models.response_user import ResponseUser
from openapi_client.models.response_user_with_token_response import ResponseUserWithTokenResponse
from openapi_client.models.response_volume import ResponseVolume
from openapi_client.models.response_volume_mount_info import ResponseVolumeMountInfo
from openapi_client.models.response_volume_mount_infos import ResponseVolumeMountInfos
from openapi_client.models.response_volume_source_dataset import ResponseVolumeSourceDataset
from openapi_client.models.response_volume_source_dataset_version import ResponseVolumeSourceDatasetVersion
from openapi_client.models.response_volume_source_volume import ResponseVolumeSourceVolume
from openapi_client.models.response_workload_endpoints import ResponseWorkloadEndpoints
from openapi_client.models.response_workload_history_info import ResponseWorkloadHistoryInfo
from openapi_client.models.response_workload_status_info import ResponseWorkloadStatusInfo
from openapi_client.models.response_workspace_backup import ResponseWorkspaceBackup
from openapi_client.models.response_workspace_detail import ResponseWorkspaceDetail
from openapi_client.models.response_workspace_list import ResponseWorkspaceList
from openapi_client.models.s3_dataset_create_api_input import S3DatasetCreateAPIInput
from openapi_client.models.ssh_key_create_api_input import SSHKeyCreateAPIInput
from openapi_client.models.service_create_api_input import ServiceCreateAPIInput
from openapi_client.models.service_service_logs_response import ServiceServiceLogsResponse
from openapi_client.models.service_service_status_running_response import ServiceServiceStatusRunningResponse
from openapi_client.models.service_service_system_metrics_response import ServiceServiceSystemMetricsResponse
from openapi_client.models.sign_in_api_input import SignInAPIInput
from openapi_client.models.sign_in_cli_confirm_api_input import SignInCliConfirmAPIInput
from openapi_client.models.sign_in_google_api_input import SignInGoogleAPIInput
from openapi_client.models.sign_up_google_api_input import SignUpGoogleAPIInput
from openapi_client.models.sign_up_pending_user_api_input import SignUpPendingUserAPIInput
from openapi_client.models.sign_up_pending_user_resolve_api_input import SignUpPendingUserResolveAPIInput
from openapi_client.models.sign_up_validate_email_api_input import SignUpValidateEmailAPIInput
from openapi_client.models.sign_up_validate_username_api_input import SignUpValidateUsernameAPIInput
from openapi_client.models.slack_authorize_api_input import SlackAuthorizeAPIInput
from openapi_client.models.storage_federation_token import StorageFederationToken
from openapi_client.models.storage_federation_token_legacy import StorageFederationTokenLegacy
from openapi_client.models.storage_file import StorageFile
from openapi_client.models.storage_file_action_url_info_legacy import StorageFileActionURLInfoLegacy
from openapi_client.models.storage_google_storage_federation_token import StorageGoogleStorageFederationToken
from openapi_client.models.storage_s3_federation_token import StorageS3FederationToken
from openapi_client.models.storage_total_size import StorageTotalSize
from openapi_client.models.sweep_create_api_input import SweepCreateAPIInput
from openapi_client.models.sweep_sweep_list_response import SweepSweepListResponse
from openapi_client.models.sweep_sweep_logs_response import SweepSweepLogsResponse
from openapi_client.models.sweep_sweep_plots_response import SweepSweepPlotsResponse
from openapi_client.models.sweep_sweep_system_metrics_response import SweepSweepSystemMetricsResponse
from openapi_client.models.sweep_update_api_input import SweepUpdateAPIInput
from openapi_client.models.tag_create_api_input import TagCreateAPIInput
from openapi_client.models.tag_tag_list_response import TagTagListResponse
from openapi_client.models.tag_update_api_input import TagUpdateAPIInput
from openapi_client.models.tutorial_update_api_input import TutorialUpdateAPIInput
from openapi_client.models.update_notification_config_api_input import UpdateNotificationConfigAPIInput
from openapi_client.models.v1_node_selector_requirement import V1NodeSelectorRequirement
from openapi_client.models.v1_toleration import V1Toleration
from openapi_client.models.vessl_dataset_create_api_input import VESSLDatasetCreateAPIInput
from openapi_client.models.verify_email_api_input import VerifyEmailAPIInput
from openapi_client.models.verify_email_check_api_input import VerifyEmailCheckAPIInput
from openapi_client.models.volume_federate_api200_response import VolumeFederateAPI200Response
from openapi_client.models.volume_file_copy_api_input import VolumeFileCopyAPIInput
from openapi_client.models.volume_file_create_api_input import VolumeFileCreateAPIInput
from openapi_client.models.volume_volume_file_list_response import VolumeVolumeFileListResponse
from openapi_client.models.volumeclaim_volume_claim_config_template import VolumeclaimVolumeClaimConfigTemplate
from openapi_client.models.volumetreenode_volume_tree_node import VolumetreenodeVolumeTreeNode
from openapi_client.models.workspace_create_api_input import WorkspaceCreateAPIInput
from openapi_client.models.workspace_update_api_input import WorkspaceUpdateAPIInput
from openapi_client.models.workspace_workspace_list_response import WorkspaceWorkspaceListResponse
from openapi_client.models.workspace_workspace_logs_response import WorkspaceWorkspaceLogsResponse
from openapi_client.models.workspace_workspace_status_running_response import WorkspaceWorkspaceStatusRunningResponse
from openapi_client.models.workspace_workspace_system_metrics_response import WorkspaceWorkspaceSystemMetricsResponse
from openapi_client.models.workspace_yaml_import_api_input import WorkspaceYamlImportAPIInput



import subprocess
from typing import List

from openapi_client import ResponseGitHubCodeRef
from vessl import logger
from vessl.util.downloader import Downloader


def clone_codes(code_refs: List[ResponseGitHubCodeRef]):
    for code_ref in code_refs:
        if code_ref.git_provider == "github":
            prefix = "x-access-token"
            git_provider_domain = "github.com"
        elif code_ref.git_provider == "gitlab":
            prefix = "oauth2"
            git_provider_domain = "gitlab.com"
        else:
            prefix = "x-token-auth"
            git_provider_domain = "bitbucket.org"

        if code_ref.git_provider_custom_domain is not None:
            git_provider_domain = code_ref.git_provider_custom_domain

        if code_ref.token is None or code_ref.token == "":
            git_url = f"https://{git_provider_domain}/{code_ref.git_owner}/{code_ref.git_repo}.git"
        else:
            git_url = f"https://{prefix}:{code_ref.token}@{git_provider_domain}/{code_ref.git_owner}/{code_ref.git_repo}.git"
        if code_ref.mount_path:
            dirname = code_ref.mount_path
        else:
            dirname = code_ref.git_repo

        try:
            subprocess.run(["git", "clone", git_url, dirname])
        except subprocess.CalledProcessError:
            dirname = f"vessl-{code_ref.git_repo}"
            logger.info(f"Falling back to '{dirname}'...")
            subprocess.run(["git", "clone", git_url, dirname])

        if code_ref.git_ref:
            subprocess.run(["/bin/sh", "-c", f"cd {dirname}; git reset --hard {code_ref.git_ref}"])

        if code_ref.git_diff_file:
            diff_file_path = f"/tmp/{code_ref.git_repo}.diff"
            Downloader.download(
                code_ref.git_diff_file.path, diff_file_path, code_ref.git_diff_file, quiet=False
            )
            subprocess.run(["/bin/sh", "-c", f"cd {dirname}; git apply {diff_file_path}"])



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from openapi_client.configuration import Configuration


class OrmProjectExperimentFieldEdges(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'charts': 'list[OrmProjectDashboardChart]',
        'configs': 'list[OrmProjectDashboardExperimentFieldConfig]',
        'filters': 'list[OrmProjectDashboardExperimentFilter]',
        'project': 'OrmProject',
        'sorts': 'list[OrmProjectDashboardExperimentSort]'
    }

    attribute_map = {
        'charts': 'charts',
        'configs': 'configs',
        'filters': 'filters',
        'project': 'project',
        'sorts': 'sorts'
    }

    def __init__(self, charts=None, configs=None, filters=None, project=None, sorts=None, local_vars_configuration=None):  # noqa: E501
        """OrmProjectExperimentFieldEdges - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._charts = None
        self._configs = None
        self._filters = None
        self._project = None
        self._sorts = None
        self.discriminator = None

        if charts is not None:
            self.charts = charts
        if configs is not None:
            self.configs = configs
        if filters is not None:
            self.filters = filters
        if project is not None:
            self.project = project
        if sorts is not None:
            self.sorts = sorts

    @property
    def charts(self):
        """Gets the charts of this OrmProjectExperimentFieldEdges.  # noqa: E501


        :return: The charts of this OrmProjectExperimentFieldEdges.  # noqa: E501
        :rtype: list[OrmProjectDashboardChart]
        """
        return self._charts

    @charts.setter
    def charts(self, charts):
        """Sets the charts of this OrmProjectExperimentFieldEdges.


        :param charts: The charts of this OrmProjectExperimentFieldEdges.  # noqa: E501
        :type charts: list[OrmProjectDashboardChart]
        """

        self._charts = charts

    @property
    def configs(self):
        """Gets the configs of this OrmProjectExperimentFieldEdges.  # noqa: E501


        :return: The configs of this OrmProjectExperimentFieldEdges.  # noqa: E501
        :rtype: list[OrmProjectDashboardExperimentFieldConfig]
        """
        return self._configs

    @configs.setter
    def configs(self, configs):
        """Sets the configs of this OrmProjectExperimentFieldEdges.


        :param configs: The configs of this OrmProjectExperimentFieldEdges.  # noqa: E501
        :type configs: list[OrmProjectDashboardExperimentFieldConfig]
        """

        self._configs = configs

    @property
    def filters(self):
        """Gets the filters of this OrmProjectExperimentFieldEdges.  # noqa: E501


        :return: The filters of this OrmProjectExperimentFieldEdges.  # noqa: E501
        :rtype: list[OrmProjectDashboardExperimentFilter]
        """
        return self._filters

    @filters.setter
    def filters(self, filters):
        """Sets the filters of this OrmProjectExperimentFieldEdges.


        :param filters: The filters of this OrmProjectExperimentFieldEdges.  # noqa: E501
        :type filters: list[OrmProjectDashboardExperimentFilter]
        """

        self._filters = filters

    @property
    def project(self):
        """Gets the project of this OrmProjectExperimentFieldEdges.  # noqa: E501


        :return: The project of this OrmProjectExperimentFieldEdges.  # noqa: E501
        :rtype: OrmProject
        """
        return self._project

    @project.setter
    def project(self, project):
        """Sets the project of this OrmProjectExperimentFieldEdges.


        :param project: The project of this OrmProjectExperimentFieldEdges.  # noqa: E501
        :type project: OrmProject
        """

        self._project = project

    @property
    def sorts(self):
        """Gets the sorts of this OrmProjectExperimentFieldEdges.  # noqa: E501


        :return: The sorts of this OrmProjectExperimentFieldEdges.  # noqa: E501
        :rtype: list[OrmProjectDashboardExperimentSort]
        """
        return self._sorts

    @sorts.setter
    def sorts(self, sorts):
        """Sets the sorts of this OrmProjectExperimentFieldEdges.


        :param sorts: The sorts of this OrmProjectExperimentFieldEdges.  # noqa: E501
        :type sorts: list[OrmProjectDashboardExperimentSort]
        """

        self._sorts = sorts

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, OrmProjectExperimentFieldEdges):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, OrmProjectExperimentFieldEdges):
            return True

        return self.to_dict() != other.to_dict()



import yaml
import time
from halo import Halo
import datetime
from vessl.util.constant import WEB_HOST, colors, LOGO
from typing import List, Optional, TextIO

from openapi_client.models import ResponseExperimentInfo
from openapi_client import ExperimentYamlImportAPIInput, WorkspaceYamlImportAPIInput
from vessl import __version__, vessl_api
from vessl.organization import _get_organization_name
from vessl.project import _get_project_name
from vessl.kernel_cluster import list_clusters

from vessl.experiment import read_experiment_by_id, list_experiment_logs
from vessl.workspace import read_workspace


def get_dt():
    dt = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    dt = f"{colors.GREY}[{dt}]{colors.ENDC}"
    return dt


def wrap_str(string, color="default", end="", do_print=True):
    if color == "cyan":
        wrapped = f"{get_dt()}{colors.OKCYAN}{string}{end}{colors.ENDC}"
    elif color == "green":
        wrapped = f"{get_dt()}{colors.OKGREEN}{string}{end}{colors.ENDC}"
    elif color == "red":
        wrapped = f"{get_dt()}{colors.FAIL}{string}{end}{colors.ENDC}"
    elif color == "warn":
        wrapped = f"{get_dt()}{colors.WARNING}{string}{end}{colors.ENDC}"
    else:
        wrapped = f"{get_dt()}{string}{end}"
    if do_print:
        print(wrapped)
    else:
        return wrapped


def msg_box(msg):
    indent = 1
    lines = msg.split("\n")
    space = " " * indent
    width = max(map(len, lines))
    box = f'╔{"═" * (width + indent * 2)}╗\n'  # upper_border
    box += "".join([f"║{space}{line:<{width}}{space}║\n" for line in lines])
    box += f'╚{"═" * (width + indent * 2)}╝'  # lower_border
    print(box)


def print_logs(logs: List[str]):
    timezone = datetime.datetime.now().astimezone().tzinfo
    for log in logs:
        ts = datetime.datetime.fromtimestamp(log.timestamp, tz=timezone).strftime("%H:%M:%S.%f")
        message = (
            log.message.replace("\\r", "\r")
            .replace("\\n", "\n")
            .replace("\\b", "\b")
            .replace("\\t", "\t")
            .replace("\\u001b", "\u001b")
        )
        for x in message.split("\n"):
            print(f"[{ts}] {x}")


# Check different stuffs in verify_yaml.
def verify_yaml(yaml_str, spinner):
    # replace \t to double space
    yaml_str = yaml_str.replace("\t", "  ")
    yaml_obj = yaml.safe_load(yaml_str)
    out_str = ""

    # Step 1: Check if all necessary keys exist.
    necessary_keys = [["image"], ["resources", "cluster"]]
    for keyset in necessary_keys:
        _yaml = yaml_obj
        for key in keyset:
            if key not in _yaml.keys():
                wrap_str(f" Field {key} does not exist! Please specify them in your yaml.", "red"),
                return False, False, False
            _yaml = _yaml[key]

    # Check interactive
    is_interactive = True if "interactive" in yaml_obj.keys() else False

    # Check resources
    yaml_resource = yaml_obj["resources"]
    cluster = yaml_resource["cluster"]
    # Collect possible cluster and gpus
    cluster_cands = list_clusters()
    cluster_ids = dict()
    cluster_gpus = dict()
    for e in cluster_cands:
        cluster_ids[e.name] = e.id
        cluster_gpus[e.name] = e.available_gpus

    # Verify cluster
    if cluster not in cluster_ids.keys():
        wrap_str(
            f" {cluster} cluster does not exist! Please select among {list(cluster_ids.keys())}.",
            "red",
        )
        return False, False, False
    else:
        out_str += wrap_str(f"   - 💡 Cluster Verified!", "cyan", "\n", do_print=False)

    # Verify accelerators
    if "accelerators" in yaml_resource.keys():
        accelerator = yaml_resource["accelerators"]
        accl_type, accl_num = accelerator.split(":")
        if accl_type not in cluster_gpus[cluster]:
            wrap_str(
                f" {accl_type} gpu does not exist in cluster! Please select among {cluster_gpus[cluster]}",
                "red",
            )
            return False, False, False
        else:
            out_str += wrap_str("   - 💡 Accelerator Verified!", "cyan", "\n", do_print=False)

    # Check ports if exists
    if is_interactive:
        out_str += wrap_str("   - 💡 Mode: Interactive", "cyan", do_print=False)
        ports = (
            yaml_obj["interactive"]["ports"] if "ports" in yaml_obj["interactive"].keys() else []
        )
        for port in ports:
            if str(port) in ["22", "8888"]:
                out_str += wrap_str(
                    f"   - ⚠️  WARNING: Port {str(port)} is used for ssh by default.",
                    "warn",
                    "\n",
                    do_print=False,
                )
    else:
        out_str += wrap_str("   - 💡 Mode: Non-Interactive", "cyan", do_print=False)

    return is_interactive, out_str, yaml_obj


# Get yaml, verify yaml
def run_from_yaml(
    yaml_file: TextIO,
    **kwargs,
) -> ResponseExperimentInfo:
    body = yaml_file.read()

    wrap_str(" Start VESSL RUN! 🏃", "green")
    organization = _get_organization_name(**kwargs)
    project = _get_project_name(**kwargs)
    wrap_str(f" VESSL Organization: {organization}", "cyan")
    wrap_str(f" VESSL Project: {project}", "cyan")

    spinner = Halo(text="Verifying YAML", text_color="cyan", spinner="dots", placement="right")
    spinner.start()
    interactive, out_str, yaml_obj = verify_yaml(body, spinner)
    if yaml_obj == False:
        spinner.stop_and_persist(
            symbol="😢", text=wrap_str(" YAML Verification Failed!", "red", do_print=False)
        )
        return
    else:
        spinner.stop_and_persist(
            symbol="✅", text=wrap_str(" YAML Verified!", "green", do_print=False)
        )
    print(out_str)
    wrap_str("📍 Check your YAML!", "green")

    clean_yaml_str = yaml.dump(yaml_obj)
    msg_box(clean_yaml_str)

    # workspace run
    if interactive:
        run_interactive_from_yaml(organization, body)
    else:
        run_batch_from_yaml(organization, project, body)
    return


# Applies yaml to cluster
def apply_yaml(organization, body, project=None, is_workspace=True):
    if is_workspace:
        workload = "workspace"
    else:
        workload = "experiment"
    spinner = Halo(
        text="Applying Run Definition to Cluster",
        text_color="cyan",
        spinner="dots",
        placement="right",
    )
    spinner.start()
    if is_workspace:
        response = vessl_api.workspace_yaml_import_api(
            organization_name=organization,
            workspace_yaml_import_api_input=WorkspaceYamlImportAPIInput(
                data=body,
            ),
        )
    else:
        response = vessl_api.experiment_yaml_import_api(
            organization_name=organization,
            project_name=project,
            experiment_yaml_import_api_input=ExperimentYamlImportAPIInput(
                data=body,
            ),
        )
    spinner.stop_and_persist(
        symbol="✅", text=wrap_str(" Run Definition applied to Cluster!", "green", do_print=False)
    )
    if is_workspace:
        link = f"{WEB_HOST}/{response.organization.name}/workspaces/{response.id}"
    else:
        link = f"{WEB_HOST}/{response.organization.name}/{response.project.name}/experiments/{response.number}"
    hlink = f"\033]8;;{link}\033\\{link}\033]8;;\033\\"
    wrap_str(
        f"💡 Check your {workload} {hlink}",
        "cyan",
    )
    return response


# Check if workload have started
def check_started(response, is_workspace=True):
    if is_workspace:
        workload = "Workspace"
        workspace_id = response.id
    else:
        workload = "Experiment"
        experiment_id = response.id

    spinner = Halo(
        text="Cluster Pending ..",
        text_color="cyan",
        spinner="dots",
        placement="right",
    )
    spinner.start()
    not_started = True
    terminated = False
    while not_started and (not terminated):
        if is_workspace:
            status = read_workspace(workspace_id=workspace_id).status
        else:
            status = read_experiment_by_id(experiment_id).status
        if status != "pending":
            not_started = False
        if status in ["failed", "stopped"]:
            terminated = True

    if terminated:
        spinner.stop_and_persist(
            symbol="🏝️", text=wrap_str(f" {workload} terminated!", "green", do_print=False)
        )
        return
    spinner.stop_and_persist(
        symbol="✅", text=wrap_str(f" Start initializing cluster!", "green", do_print=False)
    )

    spinner = Halo(
        text="Cluster Initializing .. ",
        text_color="cyan",
        spinner="dots",
        placement="right",
    )
    spinner.start()
    not_started = True
    while not_started and (not terminated):
        if is_workspace:
            status = read_workspace(workspace_id=workspace_id).status
        else:
            status = read_experiment_by_id(experiment_id).status
        if status == "running":
            not_started = False
        if status in ["failed", "stopped"]:
            terminated = True
    if terminated:
        spinner.stop_and_persist(
            symbol="🏝️", text=wrap_str(f" {workload} terminated!", "green", do_print=False)
        )
        return
    spinner.stop_and_persist(
        symbol="✅", text=wrap_str(f" Initialized cluster!", "green", do_print=False)
    )
    spinner.stop_and_persist(
        symbol="✅", text=wrap_str(f" {workload} have started!", "green", do_print=False)
    )
    print(LOGO)
    wrap_str(f" VESSL RUN have succesfully launched! 🚀", "green")


def run_interactive_from_yaml(organization, body):
    response = apply_yaml(organization, body, is_workspace=True)
    check_started(response, is_workspace=True)
    endpoints = response.endpoints.manually_defined_endpoints
    for endpoint in endpoints:
        wrap_str(f"📍 Endpoint {endpoint.name}: {endpoint.endpoint}")


def run_batch_from_yaml(organization, project, body):
    response = apply_yaml(organization, body, project, is_workspace=False)
    experiment_id = response.id
    experiment_number = response.number
    check_started(response, is_workspace=False)
    wrap_str(f" Showing experiment logs from now !", "green")

    # fetch pod outputs
    experiment_finished_dt = None
    after = 0
    first_log = True
    while True:
        if (
            read_experiment_by_id(experiment_id).status not in ["pending", "running"]
            and experiment_finished_dt is None
        ):
            experiment_finished_dt = time.time()

        if experiment_finished_dt is not None and (time.time() - experiment_finished_dt) > 5:
            break

        # worker number: 0 since we do not handle distributed exps
        logs = list_experiment_logs(
            experiment_number=experiment_number,
            before=int(time.time() - 5),
            after=after,
            worker_numer=0,
        )
        # do not print first log - generated while cluster was pending.
        if not first_log:
            print_logs(logs)
        else:
            first_log = False
        if len(logs) > 0:
            after = logs[-1].timestamp + 0.000001
        time.sleep(3)



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from openapi_client.configuration import Configuration


class ClusterClusterQuotaUsageResponse(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'organization_quota': 'dict[str, int]',
        'organization_quota_usage': 'dict[str, float]',
        'user_quota': 'dict[str, int]',
        'user_quota_usage': 'dict[str, float]'
    }

    attribute_map = {
        'organization_quota': 'organization_quota',
        'organization_quota_usage': 'organization_quota_usage',
        'user_quota': 'user_quota',
        'user_quota_usage': 'user_quota_usage'
    }

    def __init__(self, organization_quota=None, organization_quota_usage=None, user_quota=None, user_quota_usage=None, local_vars_configuration=None):  # noqa: E501
        """ClusterClusterQuotaUsageResponse - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._organization_quota = None
        self._organization_quota_usage = None
        self._user_quota = None
        self._user_quota_usage = None
        self.discriminator = None

        self.organization_quota = organization_quota
        self.organization_quota_usage = organization_quota_usage
        self.user_quota = user_quota
        self.user_quota_usage = user_quota_usage

    @property
    def organization_quota(self):
        """Gets the organization_quota of this ClusterClusterQuotaUsageResponse.  # noqa: E501


        :return: The organization_quota of this ClusterClusterQuotaUsageResponse.  # noqa: E501
        :rtype: dict[str, int]
        """
        return self._organization_quota

    @organization_quota.setter
    def organization_quota(self, organization_quota):
        """Sets the organization_quota of this ClusterClusterQuotaUsageResponse.


        :param organization_quota: The organization_quota of this ClusterClusterQuotaUsageResponse.  # noqa: E501
        :type organization_quota: dict[str, int]
        """
        if self.local_vars_configuration.client_side_validation and organization_quota is None:  # noqa: E501
            raise ValueError("Invalid value for `organization_quota`, must not be `None`")  # noqa: E501

        self._organization_quota = organization_quota

    @property
    def organization_quota_usage(self):
        """Gets the organization_quota_usage of this ClusterClusterQuotaUsageResponse.  # noqa: E501


        :return: The organization_quota_usage of this ClusterClusterQuotaUsageResponse.  # noqa: E501
        :rtype: dict[str, float]
        """
        return self._organization_quota_usage

    @organization_quota_usage.setter
    def organization_quota_usage(self, organization_quota_usage):
        """Sets the organization_quota_usage of this ClusterClusterQuotaUsageResponse.


        :param organization_quota_usage: The organization_quota_usage of this ClusterClusterQuotaUsageResponse.  # noqa: E501
        :type organization_quota_usage: dict[str, float]
        """
        if self.local_vars_configuration.client_side_validation and organization_quota_usage is None:  # noqa: E501
            raise ValueError("Invalid value for `organization_quota_usage`, must not be `None`")  # noqa: E501

        self._organization_quota_usage = organization_quota_usage

    @property
    def user_quota(self):
        """Gets the user_quota of this ClusterClusterQuotaUsageResponse.  # noqa: E501


        :return: The user_quota of this ClusterClusterQuotaUsageResponse.  # noqa: E501
        :rtype: dict[str, int]
        """
        return self._user_quota

    @user_quota.setter
    def user_quota(self, user_quota):
        """Sets the user_quota of this ClusterClusterQuotaUsageResponse.


        :param user_quota: The user_quota of this ClusterClusterQuotaUsageResponse.  # noqa: E501
        :type user_quota: dict[str, int]
        """
        if self.local_vars_configuration.client_side_validation and user_quota is None:  # noqa: E501
            raise ValueError("Invalid value for `user_quota`, must not be `None`")  # noqa: E501

        self._user_quota = user_quota

    @property
    def user_quota_usage(self):
        """Gets the user_quota_usage of this ClusterClusterQuotaUsageResponse.  # noqa: E501


        :return: The user_quota_usage of this ClusterClusterQuotaUsageResponse.  # noqa: E501
        :rtype: dict[str, float]
        """
        return self._user_quota_usage

    @user_quota_usage.setter
    def user_quota_usage(self, user_quota_usage):
        """Sets the user_quota_usage of this ClusterClusterQuotaUsageResponse.


        :param user_quota_usage: The user_quota_usage of this ClusterClusterQuotaUsageResponse.  # noqa: E501
        :type user_quota_usage: dict[str, float]
        """
        if self.local_vars_configuration.client_side_validation and user_quota_usage is None:  # noqa: E501
            raise ValueError("Invalid value for `user_quota_usage`, must not be `None`")  # noqa: E501

        self._user_quota_usage = user_quota_usage

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ClusterClusterQuotaUsageResponse):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ClusterClusterQuotaUsageResponse):
            return True

        return self.to_dict() != other.to_dict()



from test.conftest import TEST_CLUSTER_NAME, TEST_IMAGE_NAME, USE_MOCK

import pytest

import vessl
from vessl.util.exception import InvalidExperimentError


@pytest.mark.skipif(USE_MOCK, reason="Does not run if mocking is used.")
class TestExperiment:
    @pytest.mark.order(index=1)
    def test_create_experiment(self):
        # Use pytest.experiment instead of self.experiment because self gets reset
        # after each test
        pytest.experiment = vessl.create_experiment(
            cluster_name=TEST_CLUSTER_NAME,  # TODO: change when CI server is relocated
            start_command="echo 'Hello world!'",
            kernel_resource_spec_name="v1.cpu-0.mem-1",
            kernel_image_url=TEST_IMAGE_NAME,
            message="test message",
            termination_protection=False,
            hyperparameters=["VAR=value"],
        )
        assert pytest.experiment.is_distributed == False

    @pytest.mark.order(index=2)
    def test_create_distributed_experiment(self):
        pytest.distributed_experiment = vessl.create_experiment(
            cluster_name=TEST_CLUSTER_NAME,  # TODO: change when CI server is relocated
            start_command="echo 'Hello world!'",
            kernel_resource_spec_name="v1.cpu-0.mem-1",
            kernel_image_url=TEST_IMAGE_NAME,
            message="test message",
            termination_protection=False,
            hyperparameters=["VAR=value"],
            worker_count=2,
            framework_type="pytorch",
        )
        assert pytest.distributed_experiment.is_distributed == True

    @pytest.mark.order(index=3)
    def test_create_local_experiment(self):
        pytest.local_experiment = vessl.experiment.create_local_experiment()

    def test_update_experiment(self):
        pytest.experiment = vessl.update_experiment(pytest.experiment.number, message="updated message")
        assert pytest.experiment.message == "updated message"

    def test_read_experiment(self):
        experiment = vessl.read_experiment(pytest.experiment.number)
        assert experiment.is_distributed == False

    def test_read_distributed_experiment(self):
        distributed_experiment = vessl.read_experiment(pytest.distributed_experiment.number)
        assert distributed_experiment.is_distributed == True

    def test_read_experiment_by_id(self):
        vessl.read_experiment_by_id(pytest.experiment.id)

    def test_list_experiments(self):
        vessl.list_experiments()

    def test_list_experiment_logs(self):
        vessl.list_experiment_logs(pytest.experiment.number)
        vessl.list_experiment_logs(pytest.distributed_experiment.number)

    def test_list_experiment_output_files(self):
        vessl.list_experiment_output_files(pytest.experiment.number)
        vessl.list_experiment_output_files(pytest.distributed_experiment.number)
        # vessl.list_experiment_logs(pytest.local_experiment.number)

    def test_download_experiment_output_files(self):
        vessl.download_experiment_output_files(pytest.experiment.number)
        vessl.download_experiment_output_files(pytest.distributed_experiment.number)
        vessl.download_experiment_output_files(pytest.local_experiment.number)

    def test_upload_experiment_output_files(self):
        with pytest.raises(InvalidExperimentError):
            # Cannot upload to managed experiment
            vessl.upload_experiment_output_files(pytest.experiment.number, "test/fixture")
        vessl.upload_experiment_output_files(pytest.local_experiment.number, "test/fixture")



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from openapi_client.configuration import Configuration


class ResponseModelServiceRevision(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'autoscaler_config': 'OrmAutoscalerConfig',
        'available_replicas': 'int',
        'cluster_node_ids': 'list[int]',
        'created_by': 'ResponseUser',
        'created_dt': 'datetime',
        'deployment_type': 'str',
        'env_vars': 'OrmEnvVars',
        'id': 'int',
        'kernel_image': 'ResponseKernelImage',
        'kernel_resource_spec': 'ResponseKernelResourceSpec',
        'message': 'str',
        'model': 'ResponseSimpleModelInfo',
        'model_service_name': 'str',
        'number': 'int',
        'ports': 'OrmWorkloadPorts',
        'service_account_name': 'str',
        'start_command': 'str',
        'status': 'str'
    }

    attribute_map = {
        'autoscaler_config': 'autoscaler_config',
        'available_replicas': 'available_replicas',
        'cluster_node_ids': 'cluster_node_ids',
        'created_by': 'created_by',
        'created_dt': 'created_dt',
        'deployment_type': 'deployment_type',
        'env_vars': 'env_vars',
        'id': 'id',
        'kernel_image': 'kernel_image',
        'kernel_resource_spec': 'kernel_resource_spec',
        'message': 'message',
        'model': 'model',
        'model_service_name': 'model_service_name',
        'number': 'number',
        'ports': 'ports',
        'service_account_name': 'service_account_name',
        'start_command': 'start_command',
        'status': 'status'
    }

    def __init__(self, autoscaler_config=None, available_replicas=None, cluster_node_ids=None, created_by=None, created_dt=None, deployment_type=None, env_vars=None, id=None, kernel_image=None, kernel_resource_spec=None, message=None, model=None, model_service_name=None, number=None, ports=None, service_account_name=None, start_command=None, status=None, local_vars_configuration=None):  # noqa: E501
        """ResponseModelServiceRevision - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._autoscaler_config = None
        self._available_replicas = None
        self._cluster_node_ids = None
        self._created_by = None
        self._created_dt = None
        self._deployment_type = None
        self._env_vars = None
        self._id = None
        self._kernel_image = None
        self._kernel_resource_spec = None
        self._message = None
        self._model = None
        self._model_service_name = None
        self._number = None
        self._ports = None
        self._service_account_name = None
        self._start_command = None
        self._status = None
        self.discriminator = None

        if autoscaler_config is not None:
            self.autoscaler_config = autoscaler_config
        self.available_replicas = available_replicas
        if cluster_node_ids is not None:
            self.cluster_node_ids = cluster_node_ids
        self.created_by = created_by
        self.created_dt = created_dt
        self.deployment_type = deployment_type
        if env_vars is not None:
            self.env_vars = env_vars
        self.id = id
        if kernel_image is not None:
            self.kernel_image = kernel_image
        if kernel_resource_spec is not None:
            self.kernel_resource_spec = kernel_resource_spec
        self.message = message
        if model is not None:
            self.model = model
        self.model_service_name = model_service_name
        self.number = number
        if ports is not None:
            self.ports = ports
        if service_account_name is not None:
            self.service_account_name = service_account_name
        if start_command is not None:
            self.start_command = start_command
        self.status = status

    @property
    def autoscaler_config(self):
        """Gets the autoscaler_config of this ResponseModelServiceRevision.  # noqa: E501


        :return: The autoscaler_config of this ResponseModelServiceRevision.  # noqa: E501
        :rtype: OrmAutoscalerConfig
        """
        return self._autoscaler_config

    @autoscaler_config.setter
    def autoscaler_config(self, autoscaler_config):
        """Sets the autoscaler_config of this ResponseModelServiceRevision.


        :param autoscaler_config: The autoscaler_config of this ResponseModelServiceRevision.  # noqa: E501
        :type autoscaler_config: OrmAutoscalerConfig
        """

        self._autoscaler_config = autoscaler_config

    @property
    def available_replicas(self):
        """Gets the available_replicas of this ResponseModelServiceRevision.  # noqa: E501


        :return: The available_replicas of this ResponseModelServiceRevision.  # noqa: E501
        :rtype: int
        """
        return self._available_replicas

    @available_replicas.setter
    def available_replicas(self, available_replicas):
        """Sets the available_replicas of this ResponseModelServiceRevision.


        :param available_replicas: The available_replicas of this ResponseModelServiceRevision.  # noqa: E501
        :type available_replicas: int
        """
        if self.local_vars_configuration.client_side_validation and available_replicas is None:  # noqa: E501
            raise ValueError("Invalid value for `available_replicas`, must not be `None`")  # noqa: E501

        self._available_replicas = available_replicas

    @property
    def cluster_node_ids(self):
        """Gets the cluster_node_ids of this ResponseModelServiceRevision.  # noqa: E501


        :return: The cluster_node_ids of this ResponseModelServiceRevision.  # noqa: E501
        :rtype: list[int]
        """
        return self._cluster_node_ids

    @cluster_node_ids.setter
    def cluster_node_ids(self, cluster_node_ids):
        """Sets the cluster_node_ids of this ResponseModelServiceRevision.


        :param cluster_node_ids: The cluster_node_ids of this ResponseModelServiceRevision.  # noqa: E501
        :type cluster_node_ids: list[int]
        """

        self._cluster_node_ids = cluster_node_ids

    @property
    def created_by(self):
        """Gets the created_by of this ResponseModelServiceRevision.  # noqa: E501


        :return: The created_by of this ResponseModelServiceRevision.  # noqa: E501
        :rtype: ResponseUser
        """
        return self._created_by

    @created_by.setter
    def created_by(self, created_by):
        """Sets the created_by of this ResponseModelServiceRevision.


        :param created_by: The created_by of this ResponseModelServiceRevision.  # noqa: E501
        :type created_by: ResponseUser
        """
        if self.local_vars_configuration.client_side_validation and created_by is None:  # noqa: E501
            raise ValueError("Invalid value for `created_by`, must not be `None`")  # noqa: E501

        self._created_by = created_by

    @property
    def created_dt(self):
        """Gets the created_dt of this ResponseModelServiceRevision.  # noqa: E501


        :return: The created_dt of this ResponseModelServiceRevision.  # noqa: E501
        :rtype: datetime
        """
        return self._created_dt

    @created_dt.setter
    def created_dt(self, created_dt):
        """Sets the created_dt of this ResponseModelServiceRevision.


        :param created_dt: The created_dt of this ResponseModelServiceRevision.  # noqa: E501
        :type created_dt: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_dt is None:  # noqa: E501
            raise ValueError("Invalid value for `created_dt`, must not be `None`")  # noqa: E501

        self._created_dt = created_dt

    @property
    def deployment_type(self):
        """Gets the deployment_type of this ResponseModelServiceRevision.  # noqa: E501


        :return: The deployment_type of this ResponseModelServiceRevision.  # noqa: E501
        :rtype: str
        """
        return self._deployment_type

    @deployment_type.setter
    def deployment_type(self, deployment_type):
        """Sets the deployment_type of this ResponseModelServiceRevision.


        :param deployment_type: The deployment_type of this ResponseModelServiceRevision.  # noqa: E501
        :type deployment_type: str
        """
        if self.local_vars_configuration.client_side_validation and deployment_type is None:  # noqa: E501
            raise ValueError("Invalid value for `deployment_type`, must not be `None`")  # noqa: E501

        self._deployment_type = deployment_type

    @property
    def env_vars(self):
        """Gets the env_vars of this ResponseModelServiceRevision.  # noqa: E501


        :return: The env_vars of this ResponseModelServiceRevision.  # noqa: E501
        :rtype: OrmEnvVars
        """
        return self._env_vars

    @env_vars.setter
    def env_vars(self, env_vars):
        """Sets the env_vars of this ResponseModelServiceRevision.


        :param env_vars: The env_vars of this ResponseModelServiceRevision.  # noqa: E501
        :type env_vars: OrmEnvVars
        """

        self._env_vars = env_vars

    @property
    def id(self):
        """Gets the id of this ResponseModelServiceRevision.  # noqa: E501


        :return: The id of this ResponseModelServiceRevision.  # noqa: E501
        :rtype: int
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this ResponseModelServiceRevision.


        :param id: The id of this ResponseModelServiceRevision.  # noqa: E501
        :type id: int
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def kernel_image(self):
        """Gets the kernel_image of this ResponseModelServiceRevision.  # noqa: E501


        :return: The kernel_image of this ResponseModelServiceRevision.  # noqa: E501
        :rtype: ResponseKernelImage
        """
        return self._kernel_image

    @kernel_image.setter
    def kernel_image(self, kernel_image):
        """Sets the kernel_image of this ResponseModelServiceRevision.


        :param kernel_image: The kernel_image of this ResponseModelServiceRevision.  # noqa: E501
        :type kernel_image: ResponseKernelImage
        """

        self._kernel_image = kernel_image

    @property
    def kernel_resource_spec(self):
        """Gets the kernel_resource_spec of this ResponseModelServiceRevision.  # noqa: E501


        :return: The kernel_resource_spec of this ResponseModelServiceRevision.  # noqa: E501
        :rtype: ResponseKernelResourceSpec
        """
        return self._kernel_resource_spec

    @kernel_resource_spec.setter
    def kernel_resource_spec(self, kernel_resource_spec):
        """Sets the kernel_resource_spec of this ResponseModelServiceRevision.


        :param kernel_resource_spec: The kernel_resource_spec of this ResponseModelServiceRevision.  # noqa: E501
        :type kernel_resource_spec: ResponseKernelResourceSpec
        """

        self._kernel_resource_spec = kernel_resource_spec

    @property
    def message(self):
        """Gets the message of this ResponseModelServiceRevision.  # noqa: E501


        :return: The message of this ResponseModelServiceRevision.  # noqa: E501
        :rtype: str
        """
        return self._message

    @message.setter
    def message(self, message):
        """Sets the message of this ResponseModelServiceRevision.


        :param message: The message of this ResponseModelServiceRevision.  # noqa: E501
        :type message: str
        """
        if self.local_vars_configuration.client_side_validation and message is None:  # noqa: E501
            raise ValueError("Invalid value for `message`, must not be `None`")  # noqa: E501

        self._message = message

    @property
    def model(self):
        """Gets the model of this ResponseModelServiceRevision.  # noqa: E501


        :return: The model of this ResponseModelServiceRevision.  # noqa: E501
        :rtype: ResponseSimpleModelInfo
        """
        return self._model

    @model.setter
    def model(self, model):
        """Sets the model of this ResponseModelServiceRevision.


        :param model: The model of this ResponseModelServiceRevision.  # noqa: E501
        :type model: ResponseSimpleModelInfo
        """

        self._model = model

    @property
    def model_service_name(self):
        """Gets the model_service_name of this ResponseModelServiceRevision.  # noqa: E501


        :return: The model_service_name of this ResponseModelServiceRevision.  # noqa: E501
        :rtype: str
        """
        return self._model_service_name

    @model_service_name.setter
    def model_service_name(self, model_service_name):
        """Sets the model_service_name of this ResponseModelServiceRevision.


        :param model_service_name: The model_service_name of this ResponseModelServiceRevision.  # noqa: E501
        :type model_service_name: str
        """
        if self.local_vars_configuration.client_side_validation and model_service_name is None:  # noqa: E501
            raise ValueError("Invalid value for `model_service_name`, must not be `None`")  # noqa: E501

        self._model_service_name = model_service_name

    @property
    def number(self):
        """Gets the number of this ResponseModelServiceRevision.  # noqa: E501


        :return: The number of this ResponseModelServiceRevision.  # noqa: E501
        :rtype: int
        """
        return self._number

    @number.setter
    def number(self, number):
        """Sets the number of this ResponseModelServiceRevision.


        :param number: The number of this ResponseModelServiceRevision.  # noqa: E501
        :type number: int
        """
        if self.local_vars_configuration.client_side_validation and number is None:  # noqa: E501
            raise ValueError("Invalid value for `number`, must not be `None`")  # noqa: E501

        self._number = number

    @property
    def ports(self):
        """Gets the ports of this ResponseModelServiceRevision.  # noqa: E501


        :return: The ports of this ResponseModelServiceRevision.  # noqa: E501
        :rtype: OrmWorkloadPorts
        """
        return self._ports

    @ports.setter
    def ports(self, ports):
        """Sets the ports of this ResponseModelServiceRevision.


        :param ports: The ports of this ResponseModelServiceRevision.  # noqa: E501
        :type ports: OrmWorkloadPorts
        """

        self._ports = ports

    @property
    def service_account_name(self):
        """Gets the service_account_name of this ResponseModelServiceRevision.  # noqa: E501


        :return: The service_account_name of this ResponseModelServiceRevision.  # noqa: E501
        :rtype: str
        """
        return self._service_account_name

    @service_account_name.setter
    def service_account_name(self, service_account_name):
        """Sets the service_account_name of this ResponseModelServiceRevision.


        :param service_account_name: The service_account_name of this ResponseModelServiceRevision.  # noqa: E501
        :type service_account_name: str
        """

        self._service_account_name = service_account_name

    @property
    def start_command(self):
        """Gets the start_command of this ResponseModelServiceRevision.  # noqa: E501


        :return: The start_command of this ResponseModelServiceRevision.  # noqa: E501
        :rtype: str
        """
        return self._start_command

    @start_command.setter
    def start_command(self, start_command):
        """Sets the start_command of this ResponseModelServiceRevision.


        :param start_command: The start_command of this ResponseModelServiceRevision.  # noqa: E501
        :type start_command: str
        """

        self._start_command = start_command

    @property
    def status(self):
        """Gets the status of this ResponseModelServiceRevision.  # noqa: E501


        :return: The status of this ResponseModelServiceRevision.  # noqa: E501
        :rtype: str
        """
        return self._status

    @status.setter
    def status(self, status):
        """Sets the status of this ResponseModelServiceRevision.


        :param status: The status of this ResponseModelServiceRevision.  # noqa: E501
        :type status: str
        """
        if self.local_vars_configuration.client_side_validation and status is None:  # noqa: E501
            raise ValueError("Invalid value for `status`, must not be `None`")  # noqa: E501

        self._status = status

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ResponseModelServiceRevision):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ResponseModelServiceRevision):
            return True

        return self.to_dict() != other.to_dict()
